Package: changepointGA
Type: Package
Title: Changepoint Detection via Modified Genetic Algorithms
Version: 0.1.5
Date: 2026-05-18
Authors@R: c(
           person(given = "Mo",
                  family = "Li",
                  role = c("aut", "cre"),
                  email = "mo.li@louisiana.edu"),
           person(given = "QiQi", 
                  family = "Lu", 
                  role = "aut", 
                  email = "qlu2@vcu.edu")
            )
Description: The Genetic Algorithm (GA) is used to perform changepoint analysis in time series data. The package also includes an extended island version of GA, as described in Lu, Lund, and Lee (2010, <doi:10.1214/09-AOAS289>). By mimicking the principles of natural selection and evolution, GA provides a powerful stochastic search technique for solving combinatorial optimization problems. In 'changepointGA', each chromosome represents a changepoint configuration, including the number and locations of changepoints, hyperparameters, and model parameters. The package employs genetic operators—selection, crossover, and mutation—to iteratively improve solutions based on the given fitness (objective) function. Key features of 'changepointGA' include encoding changepoint configurations in an integer format, enabling dynamic and simultaneous estimation of model hyperparameters, changepoint configurations, and associated parameters. The detailed algorithmic implementation can be found in the package vignettes and in the paper of Li (2024, <doi:10.48550/arXiv.2410.15571>).
License: MIT + file LICENSE
Depends: R (>= 4.3.0)
Imports: methods, foreach, doParallel, Rcpp, clue, stats, utils
LinkingTo: Rcpp, RcppArmadillo
SystemRequirements: C++17
URL: https://github.com/mli171/changepointGA
BugReports: https://github.com/mli171/changepointGA/issues
Suggests: knitr, rmarkdown, testthat (>= 3.0.0)
Encoding: UTF-8
VignetteBuilder: knitr
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
NeedsCompilation: yes
Packaged: 2026-05-18 11:40:47 UTC; mli171
Author: Mo Li [aut, cre],
  QiQi Lu [aut]
Maintainer: Mo Li <mo.li@louisiana.edu>
Repository: CRAN
Date/Publication: 2026-05-18 15:50:02 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-05-18 18:55:46 UTC; unix
Archs: changepointGA.so.dSYM

Package: changepointGA
Type: Package
Title: Changepoint Detection via Modified Genetic Algorithms
Version: 0.1.5
Date: 2026-05-18
Authors@R: c(
           person(given = "Mo",
                  family = "Li",
                  role = c("aut", "cre"),
                  email = "mo.li@louisiana.edu"),
           person(given = "QiQi", 
                  family = "Lu", 
                  role = "aut", 
                  email = "qlu2@vcu.edu")
            )
Description: The Genetic Algorithm (GA) is used to perform changepoint analysis in time series data. The package also includes an extended island version of GA, as described in Lu, Lund, and Lee (2010, <doi:10.1214/09-AOAS289>). By mimicking the principles of natural selection and evolution, GA provides a powerful stochastic search technique for solving combinatorial optimization problems. In 'changepointGA', each chromosome represents a changepoint configuration, including the number and locations of changepoints, hyperparameters, and model parameters. The package employs genetic operators—selection, crossover, and mutation—to iteratively improve solutions based on the given fitness (objective) function. Key features of 'changepointGA' include encoding changepoint configurations in an integer format, enabling dynamic and simultaneous estimation of model hyperparameters, changepoint configurations, and associated parameters. The detailed algorithmic implementation can be found in the package vignettes and in the paper of Li (2024, <doi:10.48550/arXiv.2410.15571>).
License: MIT + file LICENSE
Depends: R (>= 4.3.0)
Imports: methods, foreach, doParallel, Rcpp, clue, stats, utils
LinkingTo: Rcpp, RcppArmadillo
SystemRequirements: C++17
URL: https://github.com/mli171/changepointGA
BugReports: https://github.com/mli171/changepointGA/issues
Suggests: knitr, rmarkdown, testthat (>= 3.0.0)
Encoding: UTF-8
VignetteBuilder: knitr
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
NeedsCompilation: yes
Packaged: 2026-05-18 11:40:47 UTC; mli171
Author: Mo Li [aut, cre],
  QiQi Lu [aut]
Maintainer: Mo Li <mo.li@louisiana.edu>
Repository: CRAN
Date/Publication: 2026-05-18 15:50:02 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-05-18 19:12:33 UTC; unix
Archs: changepointGA.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(plot,cptga)
S3method(plot,cptgaisl)
S3method(print,summary.cptga)
S3method(print,summary.cptgaisl)
export(amoc_crossover)
export(amoc_mutation)
export(amoc_population)
export(amoc_selection)
export(arima_bic)
export(arima_bic_order_pq)
export(cpt_dist)
export(cptga)
export(cptgaisl)
export(mutation)
export(random_population)
export(select_tau)
export(selection_linear_rank)
export(ts_sim)
export(uniform_crossover)
exportClasses(cptga)
exportClasses(cptgaisl)
exportMethods(summary)
import(Rcpp)
import(doParallel)
import(foreach)
import(graphics)
import(parallel)
import(stats)
importFrom(Rcpp,evalCpp)
importFrom(Rcpp,sourceCpp)
importFrom(methods,new)
importFrom(utils,str)
importFrom(utils,tail)
useDynLib(changepointGA)
useDynLib(changepointGA, .registration = TRUE)

# changepointGA

Authors: Mo Li (mo.li@louisiana.edu), QiQi Lu (qlu2@vcu.edu)
---

## Overview

Detecting changepoints in a time series of length $N$ entails evaluating up to 
$2^{N-1}$ possible changepoint models, making exhaustive enumeration 
computationally infeasible. Genetic algorithms (GAs) provide a stochastic 
way to identify the structural changes: a population of candidate models 
evolves via selection, crossover, and mutation operators until it converges 
on one changepoint model that balances the goodness-of-fit with parsimony. 
The R package changepointGA encodes each candidate model as an integer 
chromosome vector and supports both the basic single-population model GA 
and the island model GA. Parallel computing is implemented on multi-core 
hardware to further accelerate computation. Users may supply custom fitness 
functions or genetic operators, while a user-friendly wrapper streamlines 
routine analyses. Extensive simulations demonstrate that our package runs 
significantly faster than binary-encoded GA alternatives. Additionally, 
this package can simultaneously locate changepoints and estimate their 
effects, as well as other model parameters and any integer-valued 
hyperparameters. Applications to array-based comparative genomic 
hybridization data and a century-long temperature series further 
highlight the package’s value in biological and climate research.

## Installation
You can install the version of changepointGA from CRAN:

```{r}
install.packages("changepointGA")
```

or the development version from Github:

```{r}
# install.packages("remotes")
remotes::install_github("mli171/changepointGA", build_vignettes = FALSE, force = TRUE)
```

## Changepoint Detection Only

### An example of using the GA 
```{r}
##### Stationary time series with autocorrelation
Ts = 1000
betaT = c(0.5, -0.5, 0.3) # intercept, B, D
period = 30
XMatT = cbind(rep(1, Ts), cos(2*pi*(1:Ts)/period), sin(2*pi*(1:Ts)/period))
colnames(XMatT) = c("intercept", "Bvalue", "DValue")
sigmaT = 1
phiT = c(0.5)
DeltaT = c(2, -2)
Cp.prop = c(1/4, 3/4)
CpLocT = floor(Ts*Cp.prop)

Xt = ts_sim(Ts=Ts, beta=betaT, XMat=XMatT, sigma=sigmaT, phi=phiT, theta=NULL, 
            Delta=DeltaT, CpLoc=CpLocT, seed=1234)

tim1 = Sys.time()
tmp1 = cptga(ObjFunc=arima_bic, N=Ts, XMat=XMatT, Xt=Xt)
tim2 = Sys.time()
summary(tmp1)
plot(tmp1, data=Xt)


tim2 - tim1
```

### An example of using the island-based GA 
```{r}
##### Stationary time series with autocorrelation
Ts = 1000
betaT = c(0.5) # intercept
XMatT = matrix(1, nrow=Ts, ncol=1)
colnames(XMatT) = "intercept"
sigmaT = 1
phiT = c(0.5)
DeltaT = c(2, -2)
Cp.prop = c(1/4, 3/4)
CpLocT = floor(Ts*Cp.prop)

Xt = ts_sim(Ts=Ts, beta=betaT, XMat=XMatT, sigma=sigmaT, phi=phiT, theta=NULL, 
            Delta=DeltaT, CpLoc=CpLocT, seed=1234)


## No parallel computing
tim3 = Sys.time()
tmp2 = cptgaisl(ObjFunc=arima_bic, N=Ts, XMat=XMaT, Xt=Xt)
tim4 = Sys.time()
summary(tmp2)
plot(tmp2, data=Xt)


## Parallel computing
tim5 = Sys.time()
tmp3 = cptgaisl(ObjFunc=arima_bic, N=Ts, parallel=TRUE, nCore=5, XMat=XMaT, Xt=Xt)
tim6 = Sys.time()
summary(tmp3)
plot(tmp3, data=Xt)

tim4 - tim3
tim6 - tim5
```

## Changepoint Detection + Model order selection

### An example of using the GA 
```{r}
Ts = 1000
betaT = c(0.5, -0.5, 0.3) # intercept, B, D
period = 30
XMatT = cbind(rep(1, Ts), cos(2*pi*(1:Ts)/period), sin(2*pi*(1:Ts)/period))
colnames(XMatT) = c("intercept", "Bvalue", "DValue")
sigmaT = 1
phiT = c(0.5, -0.5)
thetaT = c(0.8)
DeltaT = c(2, -2)
Cp.prop = c(1/4, 3/4)
CpLocT = floor(Ts*Cp.prop)

Xt = ts_sim(Ts=Ts, beta=betaT, XMat=XMatT, sigma=sigmaT, phi=phiT, theta=thetaT, 
            Delta=DeltaT, CpLoc=CpLocT, seed=1234)

prange = list(ar=c(0,2), ma=c(0,2))

tim1 = Sys.time()
tmp1 = cptga(ObjFunc=arima_bic_order, N=Ts, prange=prange, option="both", 
             XMat=XMatT, Xt=Xt)
tim2 = Sys.time()
summary(tmp1)
plot(tmp1, data=Xt)

tim2 - tim1
```

### An example of using the island model GA 
```{r}
Ts = 1000
betaT = c(0.5, -0.5, 0.3) # intercept, B, D
period = 30
XMatT = cbind(rep(1, Ts), cos(2*pi*(1:Ts)/period), sin(2*pi*(1:Ts)/period))
colnames(XMatT) = c("intercept", "Bvalue", "DValue")
sigmaT = 1
phiT = c(0.5, -0.5)
thetaT = c(0.8)
DeltaT = c(2, -2)
Cp.prop = c(1/4, 3/4)
CpLocT = floor(Ts*Cp.prop)

myts = ts_sim(Ts=Ts, beta=betaT, XMat=XMatT, sigma=sigmaT, phi=phiT, theta=thetaT, 
              Delta=DeltaT, CpLoc=CpLocT, seed=1234)

prange = list(ar=c(0,2), ma=c(0,2))

tim3 = Sys.time()
tmp2 = cptgaisl(ObjFunc=arima_bic_order, N=Ts, prange=prange, option="both", 
                XMat=XMatT, Xt=Xt)
tim4 = Sys.time()
summary(tmp2)
plot(tmp2, data=Xt)


tim5 = Sys.time()
tmp3 = cptgaisl(ObjFunc=arima_bic_order, N=Ts, prange=prange, option="both", 
                parallel=TRUE, nCore=5, XMat=XMatT, Xt=Xt)
tim6 = Sys.time()
summary(tmp3)
plot(tmp3, data=Xt)

tim4 - tim3
tim6 - tim5
```

# Code style

Before pushing changes, please run 

```r
styler::style_pkg()
```

to ensure your code follows the tidyverse style guide.

#' S4 Class for Island Model Genetic Algorithm-Based Changepoint Detection
#'
#' This class stores results and settings for the island model genetic algorithm (`cptgaisl`)
#' used in multiple changepoint detection and optional model order selection.
#'
#' @name cptgaisl-class
#' @title S4 Class Definition for `cptgaisl`
#'
#' @slot call The matched call that created the object.
#' @slot N The sample size of the time series.
#' @slot prange A list or NULL. Ranges for each model order parameter when \code{option = "both"}.
#' @slot popSize Integer. It represents the total number of individuals in each generation, which equal to the number of islands multiplied by the size of each island (i.e., \code{popSize = numIslands × Islandsize}).
#' @slot numIslands Integer. The number of islands (sub-populations).
#' @slot Islandsize Numerical value. The number of individuals in each island.
#' @slot pcrossover Probability of crossover operation.
#' @slot pmutation Probability of mutation operation.
#' @slot pchangepoint Prior probability of changepoint occurrence.
#' @slot minDist Minimum distance between adjacent changepoints.
#' @slot mmax Maximum number of changepoints allowed.
#' @slot lmax Maximum length of chromosome.
#' @slot maxMig Maximum number of migrations allowed.
#' @slot maxgen Maximum number of generations per island before migration.
#' @slot maxconv Number of migrations with no improvement before stopping.
#' @slot option Either "cp" or "both".
#' @slot monitoring Logical. If TRUE, prints intermediate output.
#' @slot parallel Logical. Whether parallel computation is used.
#' @slot nCore Integer or NULL. Number of cores for parallelization.
#' @slot tol Tolerance threshold for fitness improvement.
#' @slot seed Integer or NULL. Seed for reproducibility.
#' @slot suggestions A list or NULL. Suggested changepoint configurations.
#' @slot Island A 3D array storing all individual chromosomes in current generation across islands. Dimensions are \code{lmax} × \code{Islandsize} × \code{numIslands}, representing chromosome length, individuals per island, and number of islands, respectively.
#' @slot IslandFit A matrix of fitness values in current generation with dimensions \code{Islandsize} × \code{numIslands}, where each column corresponds to one island's population.
#' @slot overbestchrom A vector. The best chromosome ever found.
#' @slot overbestfit Numeric. The best fitness score obtained.
#' @slot bestfit Numeric vector recording best fitness per migration.
#' @slot countMig Integer vector tracking number of migrations.
#' @slot count Integer vector tracking total generations.
#' @slot convg Integer vector for convergence diagnostics.
#'
#' @return An object of class \code{cptgaisl}
#' @seealso \code{\link{cptgaisl}}, \code{\link{cptgaisl-class}}, \code{\link{random_population}}, \code{\link{selection_linear_rank}}, \code{\link{uniform_crossover}}, \code{\link{mutation}}.

setClassUnion("numericOrNULL", members = c("numeric", "NULL"))
setClassUnion("listOrNULL", members = c("list", "NULL"))

#' @rdname cptgaisl-class
#' @export
setClass("cptgaisl",
  representation(
    call = "language",
    N = "numeric",
    prange = "listOrNULL",
    popSize = "numeric",
    numIslands = "numeric",
    Islandsize = "numeric",
    pcrossover = "numeric",
    pmutation = "numeric",
    pchangepoint = "numeric",
    minDist = "numeric",
    mmax = "numeric",
    lmax = "numeric",
    maxMig = "numeric",
    maxgen = "numeric",
    maxconv = "numeric",
    option = "character",
    monitoring = "logical",
    parallel = "logical",
    nCore = "numericOrNULL",
    tol = "numeric",
    seed = "numericOrNULL",
    suggestions = "listOrNULL",
    Island = "array",
    IslandFit = "matrix",
    overbestchrom = "vector",
    overbestfit = "numeric",
    bestfit = "vector",
    countMig = "numeric",
    count = "numeric",
    convg = "numeric"
  ),
  package = "changepointGA"
)

setMethod("print", "cptgaisl", function(x, ...) str(x))


#' Print Summary for a `cptgaisl` Object
#'
#' Displays key information about the settings and results from a changepoint detection
#' procedure using the Island model Genetic Algorithm (IMGA) stored in a `cptgaisl` object. This includes
#' the algorithm configuration, population settings, optimization mode, and final
#' solution such as the number and location of changepoints and model parameters (if applicable).
#'
#' @param x An object of class \code{cptgaisl}, typically produced by a GA-based changepoint detection routine.
#' @param digits Number of digits to print for probabilities and fitness. Default taken from \code{getOption("digits")}.
#' @param max_display Maximum number of suggested solutions to display if suggestions are provided.
#' @param ... Additional arguments (currently not used).
#'
#' @details
#' When the GA is run in \code{option = "cp"} mode, only changepoint locations are shown.
#' If \code{option = "both"}, the output includes the selected model hyperparameters along
#' with changepoint locations.
#'
#' The function uses plain text output to print a formatted summary to the console. If
#' \code{x@suggestions} is provided, only up to \code{max_display} suggestions will be shown.
#'
#' @return Invisibly returns \code{NULL}. Called for its side effect of printing to the console.
#'
#' @seealso \code{\link{cptgaisl}}, \code{\link[=summary.cptgaisl]{summary}}, \code{\link{plot.cptgaisl}}
#'
#' @method print summary.cptgaisl
#' @export
#' @aliases print.summary.cptgaisl
print.summary.cptgaisl <- function(x, digits = getOption("digits"), max_display = 5, ...) {
  cat("###############################################\n")
  cat("#  Changepoint Detection via Island Model GA  #\n")
  cat("###############################################\n")

  cat("   Settings: \n")
  cat(paste("   Population size         = ", x@popSize, "\n"))
  cat(paste("   Number of Island        = ", x@numIslands, "\n"))
  cat(paste("   Island size             = ", x@Islandsize, "\n"))
  cat(paste("   Number of generations   = ", x@count, "\n"))
  cat(paste("   Number of migrations    = ", x@countMig, "\n"))
  cat(paste("   Crossover probability   = ", format(x@pcrossover, digits = digits), "\n"))
  cat(paste("   Mutation probability    = ", format(x@pmutation, digits = digits), "\n"))
  cat(paste("   Changepoint probability = ", format(x@pchangepoint, digits = digits), "\n"))
  cat(paste("   minDist                 = ", x@minDist, "\n"))
  cat(paste("   Task mode               = ", x@option, "\n"))
  cat(paste("   Parallel Usage          = ", x@parallel, "\n"))
  if (x@parallel) {
    cat(paste("   Number of thread      = ", x@nCore, "\n"))
  }
  seed_print <- if (is.null(x@seed) || length(x@seed) == 0) {
    "NULL"
  } else {
    as.character(x@seed)
  }
  cat(paste("   Seed                    = ", seed_print, "\n"))
  if (!is.null(x@suggestions)) {
    cat("   Suggestions: \n")
    for (i in seq_along(x@suggestions)) {
      cat("    ", sprintf("[%d]:", i))
      cat(x@suggestions[[i]], sep = " ")
      if (i > max_display) {
        cat("\n     ...")
        break
      }
      cat("\n")
    }
  }
  cat("\n##### Island Model GA results ##### \n")
  cat(paste("   Optimal Fitness value =", format(x@overbestfit, digits = digits), "\n"))
  cat(paste("   Optimal Solution: \n"))
  m.sol <- x@overbestchrom[1]
  cat(paste("        Number of Changepoints = ", m.sol, "\n"))
  if (x@option == "cp") {
    if (m.sol > 0) {
      tau.sol <- x@overbestchrom[2:(1 + m.sol)]
      cat(paste("        Changepoints Locations = ", paste(tau.sol, collapse = " ")), "\n")
    } else {
      cat("        Changepoints Locations = No changepoint reached optimum \n")
    }
  } else if (x@option == "both") {
    n.hyperparam <- length(x@prange)
    name.hyperparam <- names(x@prange)
    if (is.null(name.hyperparam)) {
      name.hyperparam <- paste0("Hyper.param.", 1:n.hyperparam)
    }
    if (m.sol > 0) {
      hyperparam.sol <- x@overbestchrom[2:(1 + n.hyperparam)]
      hyperparam.sol <- paste0(name.hyperparam, " = ", hyperparam.sol)
      cat("        Model hyperparameters:\n")
      for (i in seq_along(name.hyperparam)) {
        cat(paste("            ", hyperparam.sol[i]), "\n")
      }
      tau.sol <- x@overbestchrom[(2 + n.hyperparam):(1 + n.hyperparam + m.sol)]
      cat(paste("        Changepoints Locations = ", paste(tau.sol, collapse = " ")), "\n")
    } else {
      hyperparam.sol <- x@overbestchrom[2:(1 + n.hyperparam)]
      hyperparam.sol <- paste0(name.hyperparam, " = ", hyperparam.sol)
      cat("        Model hyperparameters:\n")
      for (i in seq_along(name.hyperparam)) {
        cat(paste("            ", hyperparam.sol[i]), "\n")
      }
      cat("        Changepoints Locations = No changepoint reached optimum \n")
    }
  }

  #
  invisible()
}

#' Print method for objects of class \code{cptgaisl}
#'
#' @param object An object of class \code{cptgaisl}.
#' @param ... Additional arguments (ignored).
#' @rdname cptgaisl-class
#' @aliases summary.cptgaisl
#' @export
setMethod("summary", "cptgaisl", function(object, ...) {
  print.summary.cptgaisl(object, ...)
})


#' Plot Time Series with Detected Changepoints from a `cptgaisl` Object
#'
#' This function visualizes a univariate time series along with the changepoints
#' identified by a island model genetic algorithm, as represented by a `cptgaisl` object.
#' Vertical dashed lines mark changepoint locations, and segment means are shown as horizontal
#' dashed lines. The optimal fitness value and changepoint locations are
#' displayed as margin text.
#'
#' @param x An object of class \code{cptgaisl}, typically returned by an island model genetic algorithm based
#' changepoint detection procedure.
#' @param data A numeric vector representing the observed univariate time series.
#' @param main Optional main title for the plot.
#' @param show_segmean Binary, whether to include the segments' means.
#' @param XTickLab Optional vector (e.g., numeric or date) for custom x-axis labels.
#'        Must be the same length as \code{data}.
#' @param XTickPos Optional vector specifying which elements of \code{XTickLab} to show as ticks.
#' @param XAxisLab Optional label for the x-axis. Default is \code{"Time"}.
#' @param YAxisLab Optional label for the y-axis. Default is \code{"Data"}.
#' @param cex.lab Text size for axis labels and margin text. Default is \code{1.3}.
#' @param cex.axis Text size for axis tick labels. Default is \code{1.3}.
#' @param cex.main Text size for the main title. Default is \code{1.3}.
#' @param lwd Line width for vertical and horizontal dashed lines. Default is \code{2}.
#' @param ... Additional graphical parameters passed to \code{plot()}.
#'
#' @details
#' If \code{XTickLab} is supplied and matches the length of \code{data}, it is used for
#' the x-axis; otherwise, the default sequence \code{1:length(data)} is used.
#'
#' If the genetic algorithm was run with \code{option = "both"}, the function skips hyperparameters
#' in the chromosome when extracting changepoint positions.
#'
#' The plot displays vertical dashed lines at changepoint locations and horizontal dashed
#' lines for the mean of each segment. Fitness and changepoint summaries are shown above the plot.
#'
#' @return This function is called for its side effects and returns \code{NULL} invisibly.
#'
#' @seealso \code{\link[=summary,cptgaisl-method]{summary}}, \code{\link{print.summary.cptgaisl}}
#'
#' @exportS3Method
plot.cptgaisl <- function(x,
                          data,
                          show_segmean = TRUE,
                          main = NULL,
                          XTickLab = NULL,
                          XTickPos = NULL,
                          XAxisLab = "Time",
                          YAxisLab = "Data",
                          cex.lab = 1.3,
                          cex.axis = 1.3,
                          cex.main = 1.3,
                          lwd = 2, ...) {
  Ts <- length(data)
  use_custom_X <- !is.null(XTickLab) && length(XTickLab) == Ts
  plot_x <- if (use_custom_X) XTickLab else 1:Ts
  
  chrom <- x@overbestchrom
  m <- chrom[1]
  
  if (m > 0) {
    tau <- if (x@option == "both") {
      n.hyparam <- length(x@prange)
      chrom[(2 + n.hyparam):(1 + n.hyparam + m)]
    } else {
      chrom[2:(1 + m)]
    }
    tau <- sort(unique(tau))
  } else {
    tau <- integer(0)
  }
  
  fit <- sprintf("%.3f", x@overbestfit)
  
  starts <- c(1, tau + 1)
  ends   <- c(tau, Ts)
  
  ## changepoint labels
  if (length(tau) > 0) {
    tau_vals <- if (use_custom_X) XTickLab[tau] else tau
    changepoint_str <- paste0("Changepoints: ", paste(tau_vals, collapse = ", "))
  } else {
    changepoint_str <- "Changepoint Locations: None"
  }
  
  op <- par(c("mar", "cex.lab", "cex.axis", "cex.main"))
  on.exit(par(op), add = TRUE)
  
  par(
    mar = c(5, 5, 6, 2),
    cex.lab = cex.lab,
    cex.axis = cex.axis,
    cex.main = cex.main
  )
  
  plot(plot_x, data,
       type = "l",
       xlab = XAxisLab,
       ylab = YAxisLab,
       xaxt = "n",
       ...)
  
  if (!is.null(main)) {
    title(main = main, line = 3.5)
  }
  
  if (!is.null(XTickPos) && use_custom_X) {
    axis(1, at = XTickPos, labels = XTickPos)
  } else {
    axis(1, at = pretty(plot_x), labels = pretty(plot_x))
  }
  
  if (length(tau) > 0) {
    cp_x <- if (use_custom_X) XTickLab[tau] else tau
    abline(v = cp_x, col = "blue", lty = "dashed", lwd = lwd)
  }
  
  if (show_segmean) {
    mu.seg <- sapply(seq_along(starts), function(i) {
      mean(data[starts[i]:ends[i]])
    })
    
    for (i in seq_along(mu.seg)) {
      x0 <- if (use_custom_X) XTickLab[starts[i]] else starts[i]
      x1 <- if (use_custom_X) XTickLab[ends[i]]   else ends[i]
      
      segments(x0 = x0, y0 = mu.seg[i],
               x1 = x1, y1 = mu.seg[i],
               col = "red", lty = "dashed", lwd = lwd)
    }
  }
  
  mtext(paste("Fitness:", fit), side = 3, line = 1.5, adj = 0, cex = cex.lab)
  mtext(changepoint_str, side = 3, line = 0.5, adj = 0, cex = cex.lab)
}

#-------------------------- Check Convergence
check_conv <- function(a, maxconv, tol) {
  a <- tail(a, maxconv)
  
  if (length(a) < 2L) {
    return(0L)
  }
  if (anyNA(a) || any(!is.finite(a))) {
    return(0L)
  }
  
  as.integer(all(abs(diff(a)) < tol))
}

#' The default mutation operator in genetic algorithm
#'
#' In a certain probability, the \code{mutation} genetic operator can be applied
#' to generate a new \code{child}. By default, the new child selection can be
#' down by the similar individual selection method in population initialization,
#' \code{\link{select_tau}}.
#'
#' @param child The child chromosome resulting from the \code{crossover} genetic
#' operator.
#' @param prange Default is \code{NULL} for only changepoint detection. If
#' \code{prange} is specified as a list object, which contains the range of
#' each model order parameters for order selection (integers). The number of
#' order parameters must be equal to the length of \code{prange}.
#' @param minDist The required minimum distance between two adjacent changepoints.
#' @param pchangepoint The probability of changepoints for every time series.
#' @param lmax The user specified maximum number of changepoints, by default,
#' as \code{N/2 - 1}.
#' @param mmax The user specified maximum length of individual chromosome,
#' by default, as \code{2+N/2 + 1}.
#' @param N The sample size of the time series.
#' @details
#' A function can apply mutation to the produced child with the specified
#' probability \code{pmutation} in \code{cptga} and
#' \code{cptgaisl}. If order selection is not requested
#' (\code{option = "cp"} in \code{cptga} and \code{cptgaisl}), the default
#' \code{\link{mutation}} operator function uses \code{select_tau} to select
#' a completely new individual with a new chromosome as the mutated child.
#' For details, see \code{\link{select_tau}}. If order selection is needed
#' (\code{option = "both"} in \code{cptga} and \code{cptgaisl}), we first
#' decide whether to keep the produced child's model order with a probability
#' of 0.5. If the child's model order is retained, the \code{select_tau}
#' function is used to select a completely new individual with a new chromosome
#' as the mutated child. If a new model order is selected from the candidate
#' model order set, there is a 0.5 probability to either select a completely new
#' individual with new changepoint locations or retain the original child's
#' changepoint locations for the mutated child. Note that the current model
#' orders in the child's chromosome are excluded from the set to avoid redundant
#' objective function evaluation. Finally, the function returns a vector
#' containing the modified chromosomes for mutated \code{child}.
#' @return The resulting child chromosome representation.
#' @import Rcpp
#' @import stats
#' @import graphics
#' @useDynLib changepointGA
#' @export
mutation <- function(child, prange = NULL, minDist, pchangepoint, lmax, mmax, N) {
  plen <- length(prange)

  if (plen > 0) {
    childMut <- matrix(0, nrow = lmax, 1)
    a1 <- runif(1)
    if (a1 > 0.5) {
      # 1. order from child
      childMut[2:(plen + 1), ] <- child[2:(plen + 1)]
      # 1.1 cpt from new
      tmpchildMut <- select_tau(
        N = N, prange = NULL, minDist = minDist, pchangepoint = pchangepoint,
        mmax = mmax, lmax = lmax
      )
      childMut[1, ] <- tmpchildMut[1]
      childMut[(plen + 2):(plen + tmpchildMut[1] + 2), ] <- tmpchildMut[2:(tmpchildMut[1] + 2)]
    } else {
      # 2. order from new
      new.prange <- rep(NA, plen)
      for (ii in 1:plen) {
        tmp.prange <- setdiff(prange[[ii]][1]:prange[[ii]][2], child[2 + ii - 1])
        new.prange[ii] <- sample(tmp.prange, 1)
      }
      childMut[2:(plen + 1), ] <- new.prange
      a2 <- runif(1)
      if (a2 > 0.5) {
        # 2.1 cpt from new
        tmpchildMut <- select_tau(
          N = N, prange = NULL, minDist = minDist, pchangepoint = pchangepoint,
          mmax = mmax, lmax = lmax
        )
        childMut[1, ] <- tmpchildMut[1]
        childMut[(plen + 2):(plen + tmpchildMut[1] + 2), ] <- tmpchildMut[2:(tmpchildMut[1] + 2)]
      } else {
        # 2.2 cpt from child
        childMut[1, ] <- child[1]
        childMut[(plen + 2):(plen + child[1] + 2), ] <- child[(plen + 2):(plen + child[1] + 2)]
      }
    }
  } else {
    tmpchildMut <- select_tau(
      N = N, prange = NULL, minDist = minDist, pchangepoint = pchangepoint,
      mmax = mmax, lmax = lmax
    )
    childMut <- tmpchildMut
  }

  return(childMut)
}

new_population_Island <- function(ObjFunc, prange, selection, crossover, mutation, pop, fit, minDist, lmax, mmax, pcrossover, pmutation, pchangepoint, maxgen, N, ...) {
  # This function is used to form new population
  # some inputs ++++++++++++++++++
  #   pop= population
  #   fit= fitness evaluated for population
  #   minDist= minimum distances between two adjacent changepoints
  #   lmax= max length of chromosome
  #   mmax= max number of changepoints
  #   pcrossover= prob of crossover
  #   pmutation= prob of mutation
  #   pchangepoint= prob of changepoints for every time series
  #   maxgen= for each subpopulation, after maxgen then apply migration
  #   N= sample size
  #   X_hour= categorical time series
  #   XMat= Design matrix including covariate other than changepoint
  #   penalty= selection criterion to choose
  # outputs ++++++++++++++++++
  #   pop= updated population
  #   fit= updated population fitness
  #   bestfit  = currnt minimum of fitness values
  #   bestchrom = chromosome representation of the individual associated with bestfit

  plen <- length(prange)

  count <- 1
  repeat{
    # indicator for c("crossover", "mutation")
    #     flag[1]=1 indicating no cross-over
    #     flag[2]=1 indicating no mutation
    flag <- rep(0, 2)
    ##### step 2: parents selection
    parents <- selection(pop, fit)
    dad <- parents$dad
    mom <- parents$mom

    ##### step 3: crossover
    a1 <- runif(1)
    if (a1 <= pcrossover) {
      child <- crossover(mom, dad, prange, minDist, lmax, N)
    } else {
      child <- dad
      flag[1] <- 1
    }

    ## step 4-2: mutation
    a2 <- runif(1)
    if (a2 <= pmutation) {
      child <- mutation(child, prange, minDist, pchangepoint, lmax, mmax, N)
    } else {
      flag[2] <- 1
    }

    ## step 5: form new generation
    #  steady state method:
    #     replace the least fit in the current pop with child if child is better.
    flagsum <- flag[1] + flag[2]
    if (flagsum < 2) {
      # flagsum < 2 indicating new individual produced and fitness evaluation needed
      fitChild <- do.call(ObjFunc, c(list(child[1:(child[1] + plen + 2)], plen, ...)))
      # fitChild = do.call(ObjFunc, c(list(child[1:(child[1]+plen+2)], plen, XMat, Xt)))
      leastfit <- max(fit) # with largest fitness value

      if (fitChild < leastfit) {
        # indicating child is better than the worst one and replace
        pp <- which.max(fit)
        pop[, pp] <- child
        fit[pp] <- fitChild
      }
      count <- count + 1
    }

    # check: after every maxgen generations, apply migration in GA.Main()
    if (count >= maxgen) {
      break
    }
  }

  return(rbind(fit, pop))
}

#' Island model based genetic algorithm
#'
#' Perform the modified island-based genetic algorithm (IslandGA) for multiple changepoint detection.
#' Minimization of an objective function using genetic algorithm (GA).
#' The algorithm can be run sequentially or in explicit parallelisation.
#'
#' For any pre-specified time series model with a specified set of changepoint locations,
#' model fit is evaluated using a fitness function \eqn{Q(\boldsymbol{\theta})},
#' where \eqn{\boldsymbol{\theta}=(\boldsymbol{s},\boldsymbol{\tau},\boldsymbol{\beta})'}
#' denotes the full parameter vector. Here, \eqn{\boldsymbol{s}} denotes the set
#' of model hyperparameters, potentially including the AR or MA orders, the degree
#' of ARIMA differencing, or the periodic AR order for PAR models. The vector
#' \eqn{\boldsymbol{\tau}=\{\tau_{1}, \ldots, \tau_{m}\}} specifies the
#' changepoint locations, with the number of changepoints \eqn{m} inferred as
#' part of the estimation. Each individual chromosome representation is specified as a vector,
#' \deqn{C = (m, \boldsymbol{s}, \boldsymbol{\tau}, N+1)',}
#' where \eqn{m} represents the number of changepoints and is also equivalent to
#' the length of vector \eqn{\boldsymbol{\tau}} containing all the changepoint
#' locations. \eqn{\boldsymbol{s}} contains the integer-valued parameter for time
#' series model structure specification, such as AR, MA, or PAR orders.
#' If no model order selection is desired, then \eqn{\boldsymbol{s}} is omitted
#' and the GA detects changepoints only. The changepoint locations in
#' \eqn{\boldsymbol{\tau}} are encoded as interger values between 2 and \eqn{N},
#' allowing the length of \eqn{\boldsymbol{\tau}} to vary dynamically with \eqn{m}.
#' The value \eqn{N+1} is appended at the end of \eqn{C} to serve as a delimiter
#' marking the boundary of the chromosome.
#'
#' @param ObjFunc The fitness function to be minimized. Users can specify any R or Rcpp
#' function as the fitness function, setting the input as the potential solution to
#' the optimization problem and returning a numerical value as the output/fitness.
#' Depending on the user-specified chromosome representation, the optimization task
#' can be changepoint detection only or changepoint detection plus model order selection,
#' which can be specified via the \code{option} parameter. When
#' \code{option="both"}, the list \code{prange} must be specified to give the range
#' of model orders.
#' @param N The sample size of the time series.
#' @param prange The default value is \code{NULL} for changepoint detection only task.
#' If both model order selection and changepoint detection are required, the \code{prange}
#' argument should be provided as a list. Each element in this list must specify the range
#' for a corresponding model order parameter, and the length of the list object of
#' \code{prange} must match the number of order parameters to be estimated.
#' @param popSize An integer representing the total number of individuals in each generation, which
#' equal to the number of islands multiplied by the size of each island (i.e.,
#' \code{popSize = numIslands × Islandsize}).
#' @param numIslands An integer representing the number of islands (sub-populations).
#' @param pcrossover The probability that the crossover operator will apply to
#' the two selected parents' chromosomes to produce the offspring. The typical
#' value is close to 1, with the default setting in this package being 0.95.
#' @param pmutation The probability that the mutation operator applies on one
#' individual chromosome. Similar to the natural mutation process, new genetic
#' information is introduced to the offspring chromosome with a relatively small
#' probability (close to 0), with a default value of 0.3.
#' @param pchangepoint The probability that a changepoint has occurred. User
#' could change this probability based on domain knowledge and the time series
#' length. This probability is used during population initialization and in the
#' creation of new chromosomes by the mutation operator. By default, the mutation
#' operator function generates a new individual as the mutated offspring.
#' @param minDist The minimum length between two adjacent changepoints. Default value equals to one.
#' @param mmax The maximum number of changepoints allowed in the time series data
#' corresponds to the maximum possible length of \eqn{\boldsymbol{\tau}}.
#' For a time series of length 1000 and we only want to detect the changepoint
#' (\code{option="cp"}), the default value is 499. The suggested value should be
#' based on the length of the time series. For instance, if a time series has
#' length \eqn{N}, the recommended \code{mmax} should be \eqn{N/2-1}. It is suggested to
#' add the number of model hyperparameters if both changepoint detection and
#' model order selection tasks are of-interested simultaneously
#' (\code{option="both"}).
#' @param lmax The maximum possible length of the chromosome representation.
#' For a time series of length 1000 and we only want to detect the changepoint
#' (\code{option="cp"}), the default value is 501. The suggested value should be
#' based on the length of the time series. For instance, if a time series has
#' length \eqn{N}, the recommended \code{lmax} should be \eqn{2+N/2-1}. It is suggested to
#' add the number of model hyperparameters if both changepoint detection and
#' model order selection tasks are of-interested simultaneously (\code{option="both"}).
#' @param maxMig An integer indicates the maximum number of migrations. After
#' conducting \code{maxMig} migrations, the island model GA stops.
#' @param maxgen An integer indicates the maximum number of generations that each
#' island (subpopulation) undergoes before migration. It also determines the
#' requency of migration. The migration will occur after \code{maxgen} generations
#' for each island (sub-population).
#' @param maxconv An integer value is also used for algorithm termination. If
#' the overall best-fitted value remains unchanged after \code{maxconv}
#' consecutive migrations, the island model GA will terminate.
#' @param option A character string controls the optimization task. \code{"cp"}
#' indicates the task is changepoint detection only. \code{"both"} indicates the
#' task will include both changepoint detection and model order selection.
#' @param monitoring A logical value with either \code{TRUE} or \code{FALSE},
#' indicating whether to print out summarized results (current best fitness
#' function value and its corresponding $C$) for each generation from the GA.
#' @param parallel A logical value with \code{TRUE} or \code{FALSE}, indicating
#' whether to use multiple cores for parallel computation. Different from the
#' basic GA, in this approach, each island evolves independently. New population
#' generation—including parent selection, crossover, and mutation—is processed
#' independently and in parallel for each island, further saving computation
#' time. Once the new population generation is complete, the migration operator
#' is performed sequentially.
#' @param nCore An integer indicates the number of cores utilized in parallel
#' computing. For the island model GA, the number of cores used in parallel is
#' set to the \code{numIslands} for convenience.
#' @param tol An numerical value with the default value of \code{1e-05}. The
#' tolerance level that helps evaluate whether the two iterations have the same
#' fitness value, which aids in determining GA termination.
#' @param seed An integer with the default value of \code{NULL}. An single
#' integer allows function produce reproducible results.
#' @param popInitialize A function or sourced function name character string.
#' It should be designed for initializing a population. The default population
#' initialization is random initialization with some imposed constraints. See
#' \code{\link{random_population}} for example. The function returned object is
#' a matrix, \code{pop}. The users can specified their own \code{population}
#' function. It could also be a matrix object, which contain the user specified
#' chromosome. By default, each column represents one individual chromosome.
#' See \code{\link{random_population}} for details.
#' @param suggestions A list object. Default value is \code{NULL}. Each element
#' includes better or more reasonable guessed changepoints locations, which will
#' resulting in one chromosome. If the number of provided suggestions equals to
#' \code{popSize}, all the suggestions will be used as population. If the number
#' of provided suggestions is less than \code{popSize}, the function from
#' \code{popInitialize} will generate the remaining individual chromosomes.
#' The number of provided suggestions cannot be greater than \code{popSize}.
#' Having better \code{suggestions} can help GA converges faster.
#' @param selection A function or sourced function name character string. This
#' GA operator can help select \code{mom} and \code{dad} from current generation
#' population, where \code{dad} is set to have better fit (smaller fitness
#' function values). The default for selection uses the linear rank selection
#' method. See \code{\link{selection_linear_rank}} for example. The function
#' returned object is a list contain the chromosomes for \code{mom} and
#' \code{dad}.
#' @param crossover A function or sourced function name character string. This
#' GA operator can apply crossover to the chosen parents to produce child for
#' next generation with specified probability. The default for crossover uses
#' the uniform crossover method. See \code{\link{uniform_crossover}} for details
#' in the default crossover operator. The function returned object is a vector
#' contain the chromosomes for \code{child}.
#' @param mutation A function or sourced function name character string. This
#' GA operator can apply mutation to the produced \code{child} with the
#' specified probability \code{pmutation}. See \code{\link{mutation}} for
#' details in the default mutation operator. The function returned object
#' is a vector contain \code{child} chromosome representation.
#' @param ... additional arguments that will be passed to the fitness function.
#' @return Return an object class \code{cptgaisl-class}. See
#' \code{\link{cptgaisl-class}} for a more detailed description.
#' @import stats
#' @import Rcpp
#' @import foreach
#' @import doParallel
#' @import parallel
#' @importFrom methods new
#' @importFrom utils str
#' @useDynLib changepointGA
#' @export
#' @examples
#' \donttest{
#'
#' Ts <- 1000
#' XMatT <- matrix(1, nrow = Ts, ncol = 1)
#' Xt <- ts_sim(
#'   Ts = Ts, beta = 0.5, XMat = XMatT, sigma = 1, phi = 0.5, theta = NULL,
#'   Delta = c(2, -2), CpLoc = c(250, 750), seed = 1234
#' )
#'
#' ## Multiple changepoint detection without model order selection
#'
#' # without suggestions
#' GAISLres <- cptgaisl(ObjFunc = arima_bic, N = Ts, XMat = XMatT, Xt = Xt)
#' summary(GAISLres)
#' plot(GAISLres, data = Xt)
#'
#' # with suggestions
#' suggestions <- list(NULL, 250, c(250, 500), c(250, 625), c(250, 500, 750))
#' GAISLres <- cptgaisl(ObjFunc = arima_bic, N = Ts, suggestions = suggestions, XMat = XMatT, Xt = Xt)
#' summary(GAISLres)
#' plot(GAISLres, data = Xt)
#'
#'
#' ## Multiple changepoint detection with model order selection
#'
#' prange <- list(ar = c(0, 3), ma = c(0, 3))
#'
#' # without suggestions
#' GAISLres <- cptgaisl(
#'   ObjFunc = arima_bic_order_pq, N = Ts, prange = prange,
#'   option = "both", XMat = XMatT, Xt = Xt
#' )
#' summary(GAISLres)
#' plot(GAISLres, data = Xt)
#'
#' # with suggestions
#' suggestions <- list(NULL, 250, c(250, 500), c(250, 625), c(250, 500, 750))
#' GAISLres <- cptgaisl(
#'   ObjFunc = arima_bic_order_pq, N = Ts, prange = prange,
#'   suggestions = suggestions, option = "both", XMat = XMatT, Xt = Xt
#' )
#' summary(GAISLres)
#' plot(GAISLres, data = Xt)
#' }
cptgaisl <- function(ObjFunc,
                     N,
                     prange = NULL,
                     popSize = 200,
                     numIslands = 5,
                     pcrossover = 0.95,
                     pmutation = 0.3,
                     pchangepoint = 0.01,
                     minDist = 1,
                     mmax = NULL,
                     lmax = NULL,
                     maxMig = 1000,
                     maxgen = 50,
                     maxconv = 100,
                     option = "cp",
                     monitoring = FALSE,
                     parallel = FALSE,
                     nCore = NULL,
                     tol = 1e-05,
                     seed = NULL,
                     popInitialize = "random_population",
                     suggestions = NULL,
                     selection = "selection_linear_rank",
                     crossover = "uniform_crossover",
                     mutation = "mutation",
                     ...) {
  call <- match.call()
  dots <- list(...)

  if (missing(N)) {
    stop("The sample size must be provided")
  }
  if (missing(ObjFunc) || !is.function(ObjFunc)) {
    stop("`ObjFunc` must be a function.")
  }
  
  if (is.null(mmax)) {
    mmax <- floor(N / 2 - 1)
  }
  if (is.null(lmax)) {
    lmax <- floor(2 + N / 2 - 1)
  }
  
  plen <- length(prange)

  if (pcrossover < 0 || pcrossover > 1) {
    stop("Probability of crossover must be between 0 and 1.")
  }
  if (pmutation < 0 || pmutation > 1) {
    stop("Probability of mutation must be between 0 and 1.")
  }
  if (pchangepoint < 0 || pchangepoint > 1) {
    stop("Probability of changepoint must be between 0 and 1.")
  }
  if (minDist >= N || minDist < 1) {
    stop("Minimum number of locations between two changepoints invalid.")
  }
  if (lmax < mmax + 2) {
    stop("Maximum length of chromosome needs to be larger than (maximum number of changepoints+2).")
  }
  if (option == "both" && plen == 0) {
    stop("Opt for changepoint and order search, prange must be provided.")
  }
  if (option == "cp" && plen != 0) {
    stop("Opt for changepoint search, prange must be NULL.")
  }

  # set seed for reproducibility
  if (!is.null(seed)) {
    set.seed(seed)
  }

  if (!is.function(popInitialize)) {
    popInitialize <- get(popInitialize)
  }
  if (!is.function(selection)) {
    selection <- get(selection)
  }
  if (!is.function(crossover)) {
    crossover <- get(crossover)
  }
  if (!is.function(mutation)) {
    mutation <- get(mutation)
  }

  object <- new("cptgaisl",
    call = call,
    N = N,
    prange = prange,
    popSize = popSize,
    numIslands = numIslands,
    Islandsize = double(),
    pcrossover = pcrossover,
    pmutation = pmutation,
    pchangepoint = pchangepoint,
    minDist = minDist,
    mmax = mmax,
    lmax = lmax,
    maxMig = maxMig,
    maxgen = maxgen,
    maxconv = maxconv,
    option = option,
    monitoring = monitoring,
    parallel = parallel,
    nCore = nCore,
    tol = tol,
    seed = seed,
    suggestions = suggestions,
    Island = array(),
    IslandFit = matrix(),
    overbestchrom = vector(),
    overbestfit = double(),
    bestfit = vector(),
    countMig = integer(),
    count = integer(),
    convg = integer()
  )

  ###### step 1: generate initial population for each island
  if (!is.null(suggestions)) {
    if (!is.list(suggestions)) {
      stop("`suggestions` must be a list.")
    }
    if (any(vapply(suggestions, function(x) any(is.na(x)), logical(1)))) {
      stop("NA values are not allowed in `suggestions`.")
    }
    
    n_suggs <- length(suggestions)
    suggestions_mat <- matrix(0L, nrow = lmax, ncol = n_suggs)
    
    for (i in seq_along(suggestions)) {
      idx <- suggestions[[i]]
      
      if (!is.null(idx) && any(idx <= 1 | idx > N)) {
        stop("Suggestion ", i, " is invalid.")
      }
      
      if (option == "cp") {
        if (is.null(idx)) {
          suggestions_mat[1:2, i] <- c(0L, N + 1L)
        } else {
          idx <- idx[idx > 1 & idx <= N]
          suggestions_mat[1:(length(idx) + 2L), i] <- c(length(idx), idx, N + 1L)
        }
      } else {
        hyper_param <- vapply(
          prange,
          function(x) sample(min(x):max(x), 1),
          integer(1)
        )
        
        if (is.null(idx)) {
          suggestions_mat[1:(plen + 2L), i] <- c(0L, hyper_param, N + 1L)
        } else {
          idx <- idx[idx > 1 & idx <= N]
          suggestions_mat[1:(length(idx) + plen + 2L), i] <- c(length(idx), hyper_param, idx, N + 1L)
        }
      }
    }
    
    remain_pop_size <- popSize - n_suggs
    
    if (remain_pop_size > 0) {
      init_args <- c(
        dots,
        list(
          popSize = remain_pop_size,
          prange = prange,
          N = N,
          minDist = minDist,
          pchangepoint = pchangepoint,
          mmax = mmax,
          lmax = lmax
        )
      )
      init_args <- .filter_args(init_args, popInitialize)
      pop <- do.call(popInitialize, init_args)
      pop <- cbind(pop, suggestions_mat)
    } else if (remain_pop_size == 0) {
      pop <- suggestions_mat
    } else {
      stop("The number of suggested chromosomes must not exceed `popSize`.")
    }
  } else {
    init_args <- c(
      dots,
      list(
        popSize = popSize,
        prange = prange,
        N = N,
        minDist = minDist,
        pchangepoint = pchangepoint,
        mmax = mmax,
        lmax = lmax
      )
    )
    init_args <- .filter_args(init_args, popInitialize)
    pop <- do.call(popInitialize, init_args)
  }

  Islandsize <- floor(popSize / numIslands)
  remainder <- popSize %% numIslands
  # Shuffle individual indices
  shuffled_idx <- sample.int(popSize)
  
  Island <- array(0L, dim = c(lmax, Islandsize, numIslands))
  for (k in seq_len(numIslands)) {
    idx_start <- (k - 1L) * Islandsize + 1L
    idx_end <- k * Islandsize
    idx <- shuffled_idx[idx_start:idx_end]
    Island[, , k] <- pop[, idx, drop = FALSE]
  }
  
  if (remainder > 0L) {
    warning(remainder, " individuals were unassigned. Consider adjusting `numIslands`.")
    popSize <- numIslands * Islandsize
    object@popSize <- popSize
  }

  object@numIslands <- numIslands
  object@Islandsize <- Islandsize
  
  obj_shared <- .filter_args(dots, ObjFunc)
  obj_formals <- setdiff(names(formals(ObjFunc)), "...")
  first_param <- obj_formals[1]
  has_plen <- "plen" %in% obj_formals

  eval_fitness <- function(chromosome) {
    chrom <- trim_chromosome(chromosome, plen)
    call_list <- c(
      setNames(list(chrom), first_param),
      if (has_plen) list(plen = plen) else list(),
      obj_shared
    )
    out <- do.call(ObjFunc, call_list)
    if (!is.finite(out)) Inf else out
  }
  
  get_island_best <- function(pop_mat, fit_vec) {
    best_idx <- which.min(fit_vec)
    list(
      fit = fit_vec[best_idx],
      chrom = pop_mat[, best_idx]
    )
  }
  
  ## evaluate the fitness (Parallel or NOT)
  if (parallel) {
    n_ava_core <- parallel::detectCores()
    if (is.null(nCore)) {
      nCore <- numIslands
    }
    if (nCore < 1 || nCore > n_ava_core) {
      stop("`nCore` must be between 1 and ", n_ava_core, ".")
    }
    
    doParallel::registerDoParallel(cores = nCore)
    on.exit(foreach::registerDoSEQ(), add = TRUE)
    
    IslandFit <- foreach::foreach(k = seq_len(numIslands), .combine = "cbind") %dopar% {
      vapply(seq_len(Islandsize), function(j) eval_fitness(Island[, j, k]), numeric(1))
    }
  } else {
    IslandFit <- matrix(0, nrow = Islandsize, ncol = numIslands)
    for (k in seq_len(numIslands)) {
      IslandFit[, k] <- vapply(seq_len(Islandsize), function(j) eval_fitness(Island[, j, k]), numeric(1))
    }
  }

  object@Island <- Island
  object@IslandFit <- IslandFit

  countMig <- 0L
  countMig <- 0L
  Bfit <- rep(Inf, numIslands)
  Bchrom <- matrix(0L, nrow = lmax, ncol = numIslands)
  bestfit <- rep(NA_real_, maxMig)
  bestchrom <- matrix(0L, nrow = lmax, ncol = maxMig)

  while (countMig < maxMig) {
    
    # step 2,3,4,5: select parents, crossover, mutation, new pop
    if (parallel) {
      resNewpop <- foreach::foreach(k = seq_len(numIslands)) %dopar% {
        do.call(new_population_Island, c(
          dots,
          list(
            ObjFunc = ObjFunc,
            prange = prange,
            selection = selection,
            crossover = crossover,
            mutation = mutation,
            pop = Island[, , k],
            fit = IslandFit[, k],
            minDist = minDist,
            lmax = lmax,
            mmax = mmax,
            pcrossover = pcrossover,
            pmutation = pmutation,
            pchangepoint = pchangepoint,
            maxgen = maxgen,
            N = N
          )
        ))
      }
      
      for (k in seq_len(numIslands)) {
        tmpfit <- resNewpop[[k]][1, ]
        tmppop <- resNewpop[[k]][-1, , drop = FALSE]
        tmpfit[!is.finite(tmpfit)] <- Inf
        
        Island[, , k] <- tmppop
        IslandFit[, k] <- tmpfit
        
        island_best <- get_island_best(tmppop, tmpfit)
        Bfit[k] <- island_best$fit
        Bchrom[, k] <- island_best$chrom
      }
    } else {
      for (k in seq_len(numIslands)) {
        resNewpop <- do.call(new_population_Island, c(
          dots,
          list(
            ObjFunc = ObjFunc,
            prange = prange,
            selection = selection,
            crossover = crossover,
            mutation = mutation,
            pop = Island[, , k],
            fit = IslandFit[, k],
            minDist = minDist,
            lmax = lmax,
            mmax = mmax,
            pcrossover = pcrossover,
            pmutation = pmutation,
            pchangepoint = pchangepoint,
            maxgen = maxgen,
            N = N
          )
        ))
        tmpfit <- resNewpop[1, ]
        tmppop <- resNewpop[-1, , drop = FALSE]
        tmpfit[!is.finite(tmpfit)] <- Inf
        
        Island[, , k] <- tmppop
        IslandFit[, k] <- tmpfit
        
        island_best <- get_island_best(tmppop, tmpfit)
        Bfit[k] <- island_best$fit
        Bchrom[, k] <- island_best$chrom
      }
    }
    
    # step 6: migration
    for (k in seq_len(numIslands)) {
      worst_idx <- which.max(IslandFit[, k])
      donor_pool <- setdiff(seq_len(numIslands), k)
      if (length(donor_pool) == 0L) {next}
      donor_island <- sample(donor_pool, 1)
      Island[, worst_idx, k] <- Bchrom[, donor_island]
      IslandFit[worst_idx, k] <- Bfit[donor_island]
    }
    ## recompute island bests after migration
    for (k in seq_len(numIslands)) {
      island_best <- get_island_best(Island[, , k], IslandFit[, k])
      Bfit[k] <- island_best$fit
      Bchrom[, k] <- island_best$chrom
    }
    
    # update best
    countMig <- countMig + 1L
    genbest <- which.min(Bfit)
    bestfit[countMig] <- Bfit[genbest]
    bestchrom[, countMig] <- Bchrom[, genbest]
    
    object@countMig <- countMig
    object@count <- countMig * maxgen
    object@Island <- Island
    object@IslandFit <- IslandFit
    object@bestfit <- bestfit[seq_len(countMig)]
    
    if (countMig >= maxconv) {
      tmpbest <- bestfit[(countMig - maxconv + 1L):countMig]
      decision <- check_conv(tmpbest, maxconv, tol)
      
      if (monitoring) {
        message("Decision: ", decision)
      }
      
      if (decision == 1L) {
        object@convg <- 0L
        object@overbestfit <- bestfit[countMig]
        object@overbestchrom <- trim_chromosome(bestchrom[, countMig], plen)
        return(object)
      }
    }
    
    if (monitoring) {
      current_bestfit <- bestfit[countMig]
      current_bestchrom <- trim_chromosome(bestchrom[, countMig], plen)
      
      message(
        "==== Migration ", countMig, " ====\n",
        "countMig = ", countMig, "\n",
        "overbestfit = ", current_bestfit, "\n",
        "overbestchrom = ", paste(current_bestchrom, collapse = " ")
      )
    }
  }
  
  object@convg <- 1L
  object@overbestfit <- bestfit[countMig]
  object@overbestchrom <- trim_chromosome(bestchrom[, countMig], plen)
  
  return(object)
  
}

# tests/testthat/test-cptgaisl.R

test_that("cptgaisl returns a cptgaisl object with sensible slots", {
  d <- make_demo_data()
  
  res <- cptgaisl(
    ObjFunc = arima_bic,
    N = d$N,
    XMat = d$XMatT,
    Xt = d$Xt,
    popSize = 20,
    numIslands = 2,
    maxMig = 2,
    maxgen = 3,
    maxconv = 1,
    parallel = FALSE,
    monitoring = FALSE,
    seed = 99
  )
  
  expect_s4_class(res, "cptgaisl")
  expect_equal(res@N, d$N)
  expect_equal(res@popSize, 20)
  expect_equal(res@numIslands, 2)
  expect_equal(res@Islandsize, floor(20 / 2))
  expect_true(is.numeric(res@overbestfit))
  expect_length(res@overbestfit, 1)
  expect_true(is.finite(res@overbestfit))
  expect_true(is.numeric(res@overbestchrom))
  expect_true(length(res@overbestchrom) >= 2)
})

test_that("cptgaisl is reproducible with a fixed seed", {
  d <- make_demo_data()
  
  res1 <- cptgaisl(
    ObjFunc = arima_bic,
    N = d$N,
    XMat = d$XMatT,
    Xt = d$Xt,
    popSize = 20,
    numIslands = 2,
    maxMig = 2,
    maxgen = 3,
    maxconv = 1,
    parallel = FALSE,
    monitoring = FALSE,
    seed = 123
  )
  
  res2 <- cptgaisl(
    ObjFunc = arima_bic,
    N = d$N,
    XMat = d$XMatT,
    Xt = d$Xt,
    popSize = 20,
    numIslands = 2,
    maxMig = 2,
    maxgen = 3,
    maxconv = 1,
    parallel = FALSE,
    monitoring = FALSE,
    seed = 123
  )
  
  expect_equal(res1@overbestchrom, res2@overbestchrom)
  expect_equal(res1@overbestfit, res2@overbestfit)
})

test_that("cptgaisl summary and plot methods work", {
  d <- make_demo_data()
  
  res <- cptgaisl(
    ObjFunc = arima_bic,
    N = d$N,
    XMat = d$XMatT,
    Xt = d$Xt,
    popSize = 20,
    numIslands = 2,
    maxMig = 2,
    maxgen = 3,
    maxconv = 1,
    parallel = FALSE,
    monitoring = FALSE,
    seed = 99
  )
  
  summary_out <- capture.output(summary(res))
  expect_true(length(summary_out) > 0)
  
  tf <- tempfile(fileext = ".pdf")
  grDevices::pdf(tf)
  on.exit(grDevices::dev.off(), add = TRUE)
  
  expect_no_error(plot(res, data = d$Xt))
})

#' Calculating BIC for Multiple changepoint detection with AR and MA order selection
#'
#' The objective function for changepoint search in Autoregressive
#' moving average with model order selection via Bayesian Information Criterion
#' (BIC).
#'
#' @param chromosome A vector consists of the number of changepoints, the order
#' of AR component (refers to the number of lagged terms used to model the
#' current value of a time series), the order of MA component (refers to the
#' number of lagged error terms used to model the current value of a time
#' series), the changepoint locations, and a value of time series sample size
#' plus 1 ($N+1$) indicating the end of the chromosome.
#' @param plen The number of model order parameters that need to be selected.
#' If model order selection needs to be performed simultaneously with the
#' changepoint detection task, \code{plen} should be nonzero.
#' @param XMat A matrix contains the covariates, but not includes changepoint
#' effects, for time series regression.
#' @param Xt The simulated ARMA time series from \code{ts.sim} function.
#' @return The BIC value of the objective function.
#' @import stats
#' @importFrom utils tail
#' @useDynLib changepointGA
#' @export
#' @examples
#' Ts <- 1000
#' XMatT <- matrix(1, nrow = Ts, ncol = 1)
#' Xt <- ts_sim(
#'   Ts = Ts, beta = 0.5, XMat = XMatT, sigma = 1, phi = 0.5, theta = 0.8,
#'   Delta = c(2, -2), CpLoc = c(250, 750), seed = 1234
#' )
#'
#' # one chromosome representation
#' chromosome <- c(2, 1, 1, 250, 750, 1001)
#' arima_bic_order_pq(chromosome, plen = 2, XMat = XMatT, Xt = Xt)
arima_bic_order_pq <- function(chromosome, plen = 2, XMat, Xt) {
  
  m <- as.integer(round(chromosome[1]))
  porder <- as.integer(round(chromosome[2:(plen + 1)]))
  tau <- chromosome[(plen + 2):length(chromosome)]
  N <- length(Xt)

  if (is.na(m) || m < 0) return(1e10)
  if (any(is.na(porder)) || any(porder < 0)) return(1e10)
  
  if (m == 0) {
    DesignX <- XMat
    fit <- try(arima(Xt,
      order = c(porder[1], 0, porder[2]), xreg = DesignX, include.mean = F,
      optim.control = list(maxit = 50)
    ))
    if (inherits(fit, "try-error")) {
      bic_obj <- 1e10
    } else {
      bic_obj <- BIC(fit) 
      if (!is.finite(bic_obj)) bic_obj <- 1e10
    }
  } else {
    tau <- tau[seq_len(min(m, length(tau)))]
    tau <- as.integer(round(tau))
    tau <- tau[tau >= 1 & tau < N]
    tmptau <- sort(unique(c(tau, N)))
    CpMat <- matrix(0, nrow = N, ncol = length(tmptau) - 1)
    for (i in 1:NCOL(CpMat)) {
      CpMat[(tmptau[i] + 1):tmptau[i + 1], i] <- 1
    }
    DesignX <- cbind(XMat, CpMat)
    fit <- try(arima(Xt,
      order = c(porder[1], 0, porder[2]), xreg = DesignX, include.mean = F,
      optim.control = list(maxit = 50)
    ))
    if (inherits(fit, "try-error")) {
      bic_obj <- 1e10
    } else {
      bic_obj <- BIC(fit) + m*log(N)
      if (!is.finite(bic_obj)) bic_obj <- 1e10
    }
  }

  return(bic_obj)
}

test_that("arima_bic_order returns a finite numeric scalar", {
  Ts <- 200
  XMatT <- matrix(1, nrow = Ts, ncol = 1)
  colnames(XMatT) <- "intercept"
  
  Xt <- ts_sim(
    Ts = Ts,
    beta = 0.5,
    XMat = XMatT,
    sigma = 1,
    phi = 0.5,
    theta = 0.8,
    Delta = c(2, -2),
    CpLoc = c(50, 150),
    seed = 1234
  )
  
  chromosome <- c(2, 1, 1, 50, 150, Ts + 1)
  
  out <- arima_bic_order_pq(chromosome, plen = 2, XMat = XMatT, Xt = Xt)
  
  expect_true(is.numeric(out))
  expect_length(out, 1)
  expect_true(is.finite(out))
})

test_that("arima_bic_order is deterministic for fixed input", {
  Ts <- 200
  XMatT <- matrix(1, nrow = Ts, ncol = 1)
  colnames(XMatT) <- "intercept"
  
  Xt <- ts_sim(
    Ts = Ts,
    beta = 0.5,
    XMat = XMatT,
    sigma = 1,
    phi = 0.5,
    theta = 0.8,
    Delta = c(2, -2),
    CpLoc = c(50, 150),
    seed = 1234
  )
  
  chromosome <- c(2, 1, 1, 50, 150, Ts + 1)
  
  out1 <- arima_bic_order_pq(chromosome, plen = 2, XMat = XMatT, Xt = Xt)
  out2 <- arima_bic_order_pq(chromosome, plen = 2, XMat = XMatT, Xt = Xt)
  
  expect_equal(out1, out2)
})

#' Genetic algorithm
#'
#' Perform the genetic algorithm for changepoint detection.
#' This involves the minimization of an objective function using a genetic algorithm (GA).
#' The algorithm can be run sequentially or with explicit parallelization.
#'
#' For any pre-specified time series model with a specified set of changepoint locations,
#' model fit is evaluated using a fitness function \eqn{Q(\boldsymbol{\theta})},
#' where \eqn{\boldsymbol{\theta}=(\boldsymbol{s},\boldsymbol{\tau},\boldsymbol{\beta})'}
#' denotes the full parameter vector. Here, \eqn{\boldsymbol{s}} denotes the set
#' of model hyperparameters, potentially including the AR or MA orders, the degree
#' of ARIMA differencing, or the periodic AR order for PAR models. The vector
#' \eqn{\boldsymbol{\tau}=\{\tau_{1}, \ldots, \tau_{m}\}} specifies the
#' changepoint locations, with the number of changepoints \eqn{m} inferred as
#' part of the estimation. Each individual chromosome representation is specified as a vector,
#' \deqn{C = (m, \boldsymbol{s}, \boldsymbol{\tau}, N+1)',}
#' where \eqn{m} represents the number of changepoints and is also equivalent to
#' the length of vector \eqn{\boldsymbol{\tau}} containing all the changepoint
#' locations. \eqn{\boldsymbol{s}} contains the integer-valued parameter for time
#' series model structure specification, such as AR, MA, or PAR orders.
#' If no model order selection is desired, then \eqn{\boldsymbol{s}} is omitted
#' and the GA detects changepoints only. The changepoint locations in
#' \eqn{\boldsymbol{\tau}} are encoded as interger values between 2 and \eqn{N},
#' allowing the length of \eqn{\boldsymbol{\tau}} to vary dynamically with \eqn{m}.
#' The value \eqn{N+1} is appended at the end of \eqn{C} to serve as a delimiter
#' marking the boundary of the chromosome.
#'
#' @param ObjFunc The fitness function to be minimized. Users can specify any R or Rcpp
#' function as the fitness function, setting the input as the potential solution to
#' the optimization problem and returning a numerical value as the output/fitness.
#' Depending on the user-specified chromosome representation, the optimization task
#' can be changepoint detection only or changepoint detection plus model order selection,
#' which can be specified via the \code{option} parameter. When
#' \code{option="both"}, the list \code{prange} must be specified to give the range
#' of model orders.
#' @param N The sample size of the time series.
#' @param prange The default value is \code{NULL} for changepoint detection only task.
#' If both model order selection and changepoint detection are required, the \code{prange}
#' argument should be provided as a list. Each element in this list must specify the range
#' for a corresponding model order parameter, and the length of the list object of
#' \code{prange} must match the number of order parameters to be estimated.
#' @param popSize An integer represents the number of individuals/chromosomes in each population.
#' @param pcrossover The probability that the crossover operator will apply to
#' the two selected parents' chromosomes to produce the offspring. The typical
#' value is close to 1, with the default setting in this package being 0.95.
#' @param pmutation The probability that the mutation operator applies on one
#' individual chromosome. Similar to the natural mutation process, new genetic
#' information is introduced to the offspring chromosome with a relatively small
#' probability (close to 0), with a default value of 0.3.
#' @param pchangepoint The probability that a changepoint has occurred with a default value of 0.01. User
#' could change this probability based on domain knowledge and the time series
#' length. This probability is used during population initialization and in the
#' creation of new chromosomes by the mutation operator. By default, the mutation
#' operator function generates a new individual as the mutated offspring.
#' @param minDist The minimum length between two adjacent changepoints. Default value equals to one.
#' @param mmax The maximum number of changepoints allowed in the time series data
#' corresponds to the maximum possible length of \eqn{\boldsymbol{\tau}}.
#' For a time series of length 1000 and we only want to detect the changepoint
#' (\code{option="cp"}), the default value is 499. The suggested value should be
#' based on the length of the time series. For instance, if a time series has
#' length \eqn{N}, the recommended \code{mmax} should be \eqn{N/2-1}. It is suggested to
#' add the number of model hyperparameters if both changepoint detection and
#' model order selection tasks are of-interested simultaneously
#' (\code{option="both"}).
#' @param lmax The maximum possible length of the chromosome representation.
#' For a time series of length 1000 and we only want to detect the changepoint
#' (\code{option="cp"}), the default value is 501. The suggested value should be
#' based on the length of the time series. For instance, if a time series has
#' length \eqn{N}, the recommended \code{lmax} should be \eqn{2+N/2-1}. It is
#' suggested to add the number of model hyperparameters if both changepoint detection and
#' model order selection tasks are of-interested simultaneously (\code{option="both"}).
#' @param maxgen The maximum number of generations the GA can run before the
#' search is forcibly terminated.
#' @param maxconv If the overall best fitted value doesn't change after
#' \code{maxconv} consecutive migrations, the GA algorithm stops.
#' @param option A character string controls the optimization task. \code{"cp"}
#' indicates the task is changepoint detection only. \code{"both"} indicates the
#' task will include both changepoint detection and model order selection.
#' @param monitoring A logical value with either \code{TRUE} or \code{FALSE},
#' indicating whether to print out summarized results (current best fitness
#' function value and its corresponding $C$) for each generation from the GA.
#' @param parallel A logical value with either \code{TRUE} or \code{FALSE}, indicating
#' whether to use multiple cores for parallel computation of the fitness function
#' values for individuals after population initialization.
#' @param nCore An integer with the default value of \code{NULL}. It
#' represents the number of cores used in parallel computing. The value of
#' \code{nCore} must be less than the number of physical cores available.
#' @param tol An numerical value with the default value of \code{1e-05}. The
#' tolerance level that helps evaluate whether the two iterations have the same
#' fitness value, which aids in determining GA termination.
#' @param seed An integer with the default value of \code{NULL}. An single
#' integer allows function produce reproducible results.
#' @param popInitialize A function or sourced function name character string.
#' It should be designed for initializing a population. The default population
#' initialization is random initialization with some imposed constraints. See
#' \code{\link{random_population}} for example. The function returned object is
#' a matrix, \code{pop}. The users can specified their own \code{population}
#' function. It could also be a matrix object, which contain the user specified
#' chromosome. By default, each column represents one individual chromosome.
#' See \code{\link{random_population}} for details.
#' @param suggestions A list object. Default value is \code{NULL}. Each element
#' includes better or more reasonable guessed changepoints locations, which will
#' resulting in one chromosome. If the number of provided suggestions equals to
#' \code{popSize}, all the suggestions will be used as population. If the number
#' of provided suggestions is less than \code{popSize}, the function from
#' \code{popInitialize} will generate the remaining individual chromosomes.
#' The number of provided suggestions cannot be greater than \code{popSize}.
#' Having better \code{suggestions} can help GA converges faster.
#' @param selection A function or sourced function name character string. This
#' GA operator can help select \code{mom} and \code{dad} from current generation
#' population, where \code{dad} is set to have better fit (smaller fitness
#' function values). The default for selection uses the linear rank selection
#' method. See \code{\link{selection_linear_rank}} for example. The function
#' returned object is a list contain the chromosomes for \code{mom} and
#' \code{dad}.
#' @param crossover A function or sourced function name character string. This
#' GA operator can apply crossover to the chosen parents to produce child for
#' next generation with specified probability. The default for crossover uses
#' the uniform crossover method. See \code{\link{uniform_crossover}} for details
#' in the default crossover operator. The function returned object is a vector
#' contain the chromosomes for \code{child}.
#' @param mutation A function or sourced function name character string. This
#' GA operator can apply mutation to the produced \code{child} with the
#' specified probability \code{pmutation}. See \code{\link{mutation}} for
#' details in the default mutation operator. The function returned object
#' is a vector contain \code{child} chromosome representation.
#' @param ... additional arguments that will be passed to the fitness function.
#' @return Return an object class \code{cptga-class}. See
#' \code{\link{cptga-class}} for a more detailed description.
#' @import stats
#' @import Rcpp
#' @import foreach
#' @import doParallel
#' @import parallel
#' @importFrom methods new
#' @importFrom utils str
#' @useDynLib changepointGA
#' @export
#' @examples
#' \donttest{
#'
#' Ts <- 1000
#' XMatT <- matrix(1, nrow = Ts, ncol = 1)
#' Xt <- ts_sim(
#'   Ts = Ts, beta = 0.5, XMat = XMatT, sigma = 1, phi = 0.5, theta = NULL,
#'   Delta = c(2, -2), CpLoc = c(250, 750), seed = 1234
#' )
#'
#' ## Multiple changepoint detection without model order selection
#'
#' # without suggestions
#' GAres <- cptga(ObjFunc = arima_bic, N = Ts, XMat = XMatT, Xt = Xt)
#' summary(GAres)
#' plot(GAres, data = Xt)
#'
#' # with suggestions
#' suggestions <- list(NULL, 250, c(250, 500), c(250, 625), c(250, 500, 750))
#' GAres <- cptga(ObjFunc = arima_bic, N = Ts, suggestions = suggestions, XMat = XMatT, Xt = Xt)
#' summary(GAres)
#' plot(GAres, data = Xt)
#'
#'
#' ## Multiple changepoint detection with model order selection
#'
#' prange <- list(ar = c(0, 3), ma = c(0, 3))
#'
#' # without suggestions
#' GAres <- cptga(
#'   ObjFunc = arima_bic_order_pq, N = Ts, prange = prange,
#'   option = "both", XMat = XMatT, Xt = Xt
#' )
#' summary(GAres)
#' plot(GAres, data = Xt)
#'
#' # with suggestions
#' suggestions <- list(NULL, 250, c(250, 500), c(250, 625), c(250, 500, 750))
#' GAres <- cptga(
#'   ObjFunc = arima_bic_order_pq, N = Ts, prange = prange,
#'   suggestions = suggestions, option = "both", XMat = XMatT, Xt = Xt
#' )
#' summary(GAres)
#' plot(GAres, data = Xt)
#' }
cptga <- function(ObjFunc,
                  N,
                  prange = NULL,
                  popSize = 200,
                  pcrossover = 0.95,
                  pmutation = 0.3,
                  pchangepoint = 0.01,
                  minDist = 1,
                  mmax = NULL,
                  lmax = NULL,
                  maxgen = 50000,
                  maxconv = 5000,
                  option = "cp",
                  monitoring = FALSE,
                  parallel = FALSE,
                  nCore = NULL,
                  tol = 1e-05,
                  seed = NULL,
                  popInitialize = "random_population",
                  suggestions = NULL,
                  selection = "selection_linear_rank",
                  crossover = "uniform_crossover",
                  mutation = "mutation",
                  ...) {
  call <- match.call()
  dots <- list(...)

  if (missing(N)) {
    stop("The sample size must be provided")
  }

  if (is.null(mmax)) {
    mmax <- floor(N / 2 - 1)
  }
  if (is.null(lmax)) {
    lmax <- floor(2 + N / 2 - 1)
  }

  if (missing(ObjFunc)) {
    stop("A fitness function must be provided")
  }
  if (!is.function(ObjFunc)) {
    stop("A fitness function must be provided")
  }
  
  plen <- length(prange)

  if (pcrossover < 0 || pcrossover > 1) {
    stop("Probability of crossover must be between 0 and 1.")
  }
  if (pmutation < 0 || pmutation > 1) {
    stop("Probability of mutation must be between 0 and 1.")
  }
  if (pchangepoint < 0 || pchangepoint > 1) {
    stop("Probability of changepoint must be between 0 and 1.")
  }
  if (minDist >= N || minDist < 1) {
    stop("Minimum number of locations between two changepoints invalid.")
  }
  if (lmax < mmax + 2) {
    stop("Maximum length of chromosome needs to be larger than (maximum number of changepoints+2).")
  }
  if (option == "both" && plen == 0) {
    stop("Opt for changepoint and order search, prange must be provided.")
  }
  if (option == "cp" && plen != 0) {
    stop("Opt for changepoint search, prange must be NULL.")
  }

  # set seed for reproducibility
  if (!is.null(seed)) {
    set.seed(seed)
  }
  if (!is.function(popInitialize)) {
    popInitialize <- get(popInitialize)
  }
  if (!is.function(selection)) {
    selection <- get(selection)
  }
  if (!is.function(crossover)) {
    crossover <- get(crossover)
  }
  if (!is.function(mutation)) {
    mutation <- get(mutation)
  }

  object <- new("cptga",
    call = call,
    N = N,
    prange = prange,
    popSize = popSize,
    pcrossover = pcrossover,
    pmutation = pmutation,
    pchangepoint = pchangepoint,
    minDist = minDist,
    mmax = mmax,
    lmax = lmax,
    maxgen = maxgen,
    maxconv = maxconv,
    option = option,
    monitoring = monitoring,
    parallel = parallel,
    nCore = nCore,
    tol = tol,
    seed = seed,
    suggestions = suggestions,
    population = matrix(),
    fitness = vector(),
    overbestchrom = vector(),
    overbestfit = double(),
    bestfit = vector(),
    count = integer(),
    convg = integer()
  )

  ###### step 1: Initialize population
  if (!is.null(suggestions)) {
    if (!is.list(suggestions)) {
      stop("Suggested changepoints must be a List.")
    }
    # 1. with suggestions
    if (any(sapply(suggestions, function(x) any(is.na(x))))) {
      stop("NA provided in suggestions list")
    }
    
    n_suggs <- length(suggestions)
    suggestions_mat <- matrix(0L, nrow = lmax, ncol = n_suggs)
    for (i in seq_along(suggestions)) {
      idx <- suggestions[[i]]
      if (!is.null(idx) && any(idx <= 1 | idx > N)) {
        stop("Suggestion ", i, " is invalid.")
      }
      if (option == "cp") {
        if (is.null(idx)) {
          suggestions_mat[1:2, i] <- c(0L, N + 1L)
        } else {
          suggestions_mat[1:(length(idx) + 2L), i] <- c(length(idx), idx, N + 1L)
        }
      } else {
        hyper_param <- sapply(prange, function(x) if (length(x) > 0) sample(min(x):max(x), 1) else NA)
        
        if (is.null(idx)) {
          suggestions_mat[1:(plen + 2L), i] <- c(0L, hyper_param, N + 1L)
        } else {
          suggestions_mat[1:(length(idx) + plen + 2L), i] <- c(length(idx), hyper_param, idx, N + 1L)
        }
      }
    }
    remain_pop_size <- popSize - n_suggs
    if (remain_pop_size > 0) {
      # 1.1 partial from suggestions
      init_args <- c(
        dots,
        list(
          popSize = remain_pop_size, 
          prange = prange, 
          N = N,
          minDist = minDist, 
          pchangepoint = pchangepoint,
          mmax = mmax, 
          lmax = lmax
        )
      )
      init_args <- .filter_args(init_args, popInitialize)
      pop <- do.call(popInitialize, init_args)
      pop <- cbind(pop, suggestions_mat)
    } else if (remain_pop_size == 0) {
      # 1.2 completely from suggestions
      pop <- suggestions_mat
    } else {
      # 1.3 some errors
      stop("The number of suggested chromosomes must not exceed `popSize`.")
    }
  } else {
    # 2. completely generated by popInitialize
    init_args <- c(
      dots,
      list(
        popSize = popSize, 
        prange = prange, 
        N = N,
        minDist = minDist, 
        pchangepoint = pchangepoint,
        mmax = mmax, 
        lmax = lmax
      )
    )
    init_args <- .filter_args(init_args, popInitialize)
    pop <- do.call(popInitialize, init_args)
  }

  obj_shared <- .filter_args(dots, ObjFunc)
  obj_formals <- setdiff(names(formals(ObjFunc)), "...")
  first_param <- obj_formals[1]
  has_plen <- "plen" %in% obj_formals

  ## evaluate the fitness (Parallel or NOT)
  eval_fitness <- function(chromosome) {
    chrom <- trim_chromosome(chromosome, plen)
    call_list <- c(
      setNames(list(chrom), first_param),
      if (has_plen) list(plen = plen) else list(),
      obj_shared
    )
    out <- do.call(ObjFunc, call_list)
    if (!is.finite(out)) Inf else out
  }
  
  ## initial fitness evaluation
  if (parallel) {
    n_ava_core <- parallel::detectCores()
    
    if (is.null(nCore)) {
      stop("`nCore` must be provided when `parallel = TRUE` (", n_ava_core, " cores available).")
    }
    if (nCore < 1 || nCore > n_ava_core) {
      stop("`nCore` must be between 1 and ", n_ava_core, ".")
    }
    
    doParallel::registerDoParallel(cores = nCore)
    on.exit(foreach::registerDoSEQ(), add = TRUE)
    
    popFit <- foreach::foreach(j = seq_len(popSize), .combine = "c") %dopar% {
      eval_fitness(pop[, j])
    }
  } else {
    popFit <- vapply(seq_len(popSize), function(j) eval_fitness(pop[, j]), numeric(1))
  }

  sel_args <- .filter_args(dots, selection)
  cross_args <- .filter_args(dots, crossover)
  mut_args <- .filter_args(dots, mutation)
  
  count <- 0L
  bestfit <- rep(NA_real_, maxgen)
  bestchrom <- matrix(0, nrow = lmax, ncol = maxgen)
  
  while (count < maxgen) {
    # indicator for c("crossover", "mutation")
    #   flag[1] = 1 indicating no crossover
    #   flag[2] = 1 indicating no mutation
    flag <- c(0L, 0L)
    
    ##### step 2: parents selection
    parents <- do.call(selection, c(list(pop, popFit), sel_args))
    dad <- parents$dad
    mom <- parents$mom
    
    ##### step 3: crossover
    if (runif(1) <= pcrossover) {
      child <- do.call(crossover, c(
        list(
          mom = mom, 
          dad = dad,
          prange = prange, 
          minDist = minDist,
          lmax = lmax, 
          N = N
        ),
        cross_args
      ))
    } else {
      child <- dad
      flag[1] <- 1L
    }
    
    ##### step 4: mutation
    if (runif(1) <= pmutation) {
      child <- do.call(mutation, c(
        list(
          child, 
          prange, 
          minDist,
          pchangepoint, 
          lmax, 
          mmax, 
          N
        ),
        mut_args
      ))
    } else {
      flag[2] <- 1L
    }
    
    ##### step 5: form new generation
    if (sum(flag) < 2L) {
      fit_child <- eval_fitness(child)
      worst_idx <- which.max(popFit)
      
      if (fit_child < popFit[worst_idx]) {
        pop[, worst_idx] <- child
        popFit[worst_idx] <- fit_child
      }
    }
    
    count <- count + 1L
    genbest <- which.min(popFit)
    bestfit[count] <- popFit[genbest]
    bestchrom[, count] <- pop[, genbest]
    
    current_bestfit <- bestfit[count]
    current_bestchrom <- trim_chromosome(bestchrom[, count], plen)
    
    if (monitoring) {
      message(
        "==== Generation ", count, " ====\n",
        "overall bestfit = ", current_bestfit, "\n",
        "overall bestchrom = ", paste(current_bestchrom, collapse = " ")
      )
    }
    
    # check convergence once count >= maxconv
    # if after maxconv consecutive generations, the overall best does not change, then stop
    if (count >= maxconv) {
      tmpbestfit <- bestfit[(count - maxconv + 1L):count]
      decision <- check_conv(tmpbestfit, maxconv, tol)
      
      if (monitoring) {
        message("Decision: ", decision)
      }
      
      if (decision == 1L) {
        object@population <- pop
        object@fitness <- popFit
        object@count <- count
        object@convg <- 0L
        object@bestfit <- bestfit[seq_len(count)]
        object@overbestfit <- current_bestfit
        object@overbestchrom <- current_bestchrom
        return(object)
      }
    }
  }
  
  object@population <- pop
  object@fitness <- popFit
  object@count <- count
  object@convg <- 1L
  object@bestfit <- bestfit[seq_len(count)]
  object@overbestfit <- bestfit[count]
  object@overbestchrom <- trim_chromosome(bestchrom[, count], plen)

  return(object)
}

test_that("cptga returns a cptga object with sensible slots", {
  d <- make_demo_data()
  
  res <- cptga(
    ObjFunc = arima_bic,
    N = d$N,
    XMat = d$XMatT,
    Xt = d$Xt,
    popSize = 12,
    maxgen = 4,
    maxconv = 2,
    parallel = FALSE,
    monitoring = FALSE,
    seed = 99
  )
  
  expect_s4_class(res, "cptga")
  expect_equal(res@N, d$N)
  expect_true(is.numeric(res@overbestfit))
  expect_length(res@overbestfit, 1)
  expect_true(is.finite(res@overbestfit))
  expect_true(is.numeric(res@bestfit))
})

test_that("cptga is reproducible with a fixed seed", {
  d <- make_demo_data()
  
  res1 <- cptga(
    ObjFunc = arima_bic,
    N = d$N,
    XMat = d$XMatT,
    Xt = d$Xt,
    popSize = 12,
    maxgen = 4,
    maxconv = 2,
    parallel = FALSE,
    monitoring = FALSE,
    seed = 123
  )
  
  res2 <- cptga(
    ObjFunc = arima_bic,
    N = d$N,
    XMat = d$XMatT,
    Xt = d$Xt,
    popSize = 12,
    maxgen = 4,
    maxconv = 2,
    parallel = FALSE,
    monitoring = FALSE,
    seed = 123
  )
  
  expect_equal(res1@overbestchrom, res2@overbestchrom)
  expect_equal(res1@overbestfit, res2@overbestfit)
})

test_that("cptga summary and plot methods work", {
  d <- make_demo_data()
  
  res <- cptga(
    ObjFunc = arima_bic,
    N = d$N,
    XMat = d$XMatT,
    Xt = d$Xt,
    popSize = 12,
    maxgen = 4,
    maxconv = 2,
    parallel = FALSE,
    monitoring = FALSE,
    seed = 99
  )
  
  s <- capture.output(summary(res))
  expect_true(length(s) > 0)
  expect_true(any(grepl("GA results|Optimal Fitness value", s)))
  
  tf <- tempfile(fileext = ".pdf")
  grDevices::pdf(tf)
  on.exit(grDevices::dev.off(), add = TRUE)
  
  expect_no_error(plot(res, data = d$Xt))
})

#' Comparing multiple changepoint configurations by pairwise distance
#'
#' This function is to calculate the distance metric (Shi et al., 2022) between
#' two changepoint configurations, \eqn{C_{1}=\{\tau_{1}, \ldots, \tau_{m}\}}
#' and \eqn{C_{2}=\{\eta_{1}, \ldots, \eta_{k}\}}, where \eqn{\tau}'s are the
#' changepoint locations.
#'
#' The pairwise distance was proposed by Shi et al. (2022),
#' \deqn{d(C_{1}, C_{2})=|m-k|+ \min(A(C_{1}, C_{2})),}
#' where \eqn{m} is the number of changepoints in configuration \eqn{C_{1}} and
#' \eqn{k} is the number of changepoints in configuration \eqn{C_{2}}.
#' The term \eqn{\min(A(C_{1}, C_{2}))} reflects the cost of matching
#' changepoint locations between \eqn{C_{1}} and \eqn{C_{2}} and can be calculated
#' using the linear assignment method. Details can be found in Shi et al. (2022).
#' Note: if one configuration doesn't contain any changepoints (valued \code{NULL}),
#' the distance is defined as \eqn{|m-k|}. The function can also be used
#' to examine changepoint detection performance in simulation studies. Given the
#' true changepoint configuration, \eqn{C_{true}}, used in generating the time
#' series, the calculated distance between the estimated multiple changepoint
#' configuration, \eqn{C_{est}=\{\eta_{1}, \ldots, \eta_{k}\}}, and \eqn{C_{true}}
#' can be used to evaluate the performance of the detection algorithm.
#' @param tau1 A vector contains the changepoint locations for \eqn{C_{1}}
#' for comparison. A value \code{NULL} is required if there is no changepoint detected.
#' @param tau2 A vector contains the changepoint locations for  \eqn{C_{2}}
#' for comparison. A value \code{NULL} is required if there is no changepoint detected.
#' @param N The simulated time series sample size. Two changepoint configurations
#' should have the same \eqn{n} values.
#' @return
#' \item{dist}{The calculated distance.}
#' @references{
#'   Shi, X., Gallagher, C., Lund, R., & Killick, R. (2022).
#'   A comparison of single and multiple changepoint techniques for time series data.
#'   \emph{Computational Statistics & Data Analysis}, 170, 107433.
#' }
#' @useDynLib changepointGA
#' @export
#' @examples
#' N <- 100
#'
#' # both tau1 and tau2 has detected changepoints
#' tau2 <- c(25, 50, 75)
#' tau1 <- c(20, 35, 70, 80, 90)
#' cpt_dist(tau1 = tau1, tau2 = tau2, N = N)
#'
#' # either tau1 or tau2 has zero detected changepoints
#' cpt_dist(tau1 = tau1, tau2 = NULL, N = N)
#' cpt_dist(tau1 = NULL, tau2 = tau2, N = N)
#'
#' # both tau1 and tau2 has zero detected changepoints
#' cpt_dist(tau1 = NULL, tau2 = NULL, N = N)
cpt_dist <- function(tau1, tau2, N) {
  m <- length(tau1)
  k <- length(tau2)

  if (is.null(tau1) & is.null(tau2)) {
    dist.res <- 0
    warning("Both configurations are NULL.")
  } else {
    if (is.null(tau1) | is.null(tau2)) {
      ACC <- 0
    } else {
      if (any(tau1 < 0 | tau1 > N)) {
        stop("First changepoint configuration invalid.")
      }
      if (any(tau2 < 0 | tau2 > N)) {
        stop("Second changepoint configuration invalid.")
      }
      costs <- outer(tau1, tau2, function(x, y) abs(x - y)) / N
      if (m > k) {
        costs <- t(costs)
      }
      y <- clue::solve_LSAP(costs, maximum = FALSE)
      ACC <- sum(costs[cbind(seq_along(y), y)])
    }
    dist.res <- abs(m - k) + ACC
  }

  return(dist.res)
}

test_that("cpt_dist is zero for identical changepoint configurations", {
  tau <- c(20, 35, 70)
  expect_equal(cpt_dist(tau1 = tau, tau2 = tau, N = 100), 0)
})

test_that("cpt_dist handles empty changepoint configurations", {
  tau <- c(25, 50, 75)
  
  expect_equal(cpt_dist(tau1 = NULL, tau2 = tau, N = 100), 3)
  expect_equal(cpt_dist(tau1 = tau, tau2 = NULL, N = 100), 3)
  expect_warning(
    expect_equal(cpt_dist(tau1 = NULL, tau2 = NULL, N = 100), 0),
    "Both configurations are NULL."
  )
})

test_that("cpt_dist is symmetric", {
  tau1 <- c(20, 35, 70, 80, 90)
  tau2 <- c(25, 50, 75)
  
  expect_equal(
    cpt_dist(tau1 = tau1, tau2 = tau2, N = 100),
    cpt_dist(tau1 = tau2, tau2 = tau1, N = 100)
  )
})

test_that("cpt_dist returns a finite numeric scalar", {
  out <- cpt_dist(tau1 = c(20, 40), tau2 = c(25, 50), N = 100)
  
  expect_true(is.numeric(out))
  expect_length(out, 1)
  expect_true(is.finite(out))
})

#' Internal: filter a dots-list to only args a function accepts
#'
#' @description
#' Returns the subset of `dots` whose names match the formal arguments of `fn`
#' (excluding `...`). Handy for routing `...` to different plugin functions.
#'
#' @param dots A named list (typically `list(...)`).
#' @param fn A function object or a character name of a function.
#'
#' @return A named list containing only the args that `fn` declares.
#' @keywords internal
#' @noRd
.filter_args <- function(dots, fn) {
  if (is.character(fn)) fn <- get(fn, mode = "function")
  stopifnot(is.function(fn))

  # guard: if dots unnamed or empty
  if (length(dots) == 0L || is.null(names(dots))) {
    return(list())
  }

  fmls <- names(formals(fn))
  fmls <- setdiff(fmls, "...")
  dots[intersect(names(dots), fmls)]
}

#' Internal: trim chromosome to its active length
#'
#' @param chromosome A chromosome vector.
#' @param plen Number of AR/MA order parameters stored before changepoints.
#'
#' @return A trimmed chromosome vector of length `chromosome[1] + plen + 2`.
#' @keywords internal
#' @noRd
trim_chromosome <- function(chromosome, plen) {
  chromosome[seq_len(chromosome[1] + plen + 2L)]
}

if (getRversion() >= "2.15.1") {
  utils::globalVariables(c("j"))
}

#' changepointGA: Genetic Algorithms for Changepoint Models
#'
#' Tools for fitting changepoint models with genetic algorithms and Armadillo-backed linear algebra.
#'
#' @keywords internal
#' @useDynLib changepointGA, .registration = TRUE
#' @importFrom Rcpp evalCpp
"_PACKAGE"

#' Random population initialization for AMOC problem
#'
#' Randomly generate the individuals' chromosomes (changepoint configurations)
#' to construct the first generation population for the at most one changepoint
#' (AMOC) problem.
#'
#' @param popSize An integer represents the number of individual in each
#' population for GA (or subpopulation for IslandGA).
#' @param prange Default is \code{NULL} for only changepoint detection. If
#' \code{prange} is specified as a list object, which contains the range of
#' each model order parameters for order selection (integers). The number of
#' order parameters must be equal to the length of \code{prange}.
#' @param N The length of time series.
#' @param minDist The minimum length between two adjacent changepoints.
#' @param pchangepoint The probability that a changepoint can occur.
#' @param mmax The maximum possible number of changepoints in the data set.
#' @param lmax The maximum possible length of the chromosome representation.
#' @details
#' Given the possible candidate changepoint location set, each chromosome in the
#' first generation population can be obtained by randomly sampling one location
#' from the candidate set. The first element of every chromosome represent the
#' number of changepoints and the last non-zero element always equal to the
#' length of time series plus one (N+1).
#' @return A matrix that contains each individual's chromosome.
#' @import Rcpp
#' @import stats
#' @import graphics
#' @useDynLib changepointGA
#' @export
amoc_population <- function(popSize, prange, N, minDist, pchangepoint, mmax, lmax) {
  tauclc <- floor(0.05 * N):ceiling(0.95 * N)

  pop <- matrix(0, nrow = lmax, ncol = popSize)
  pop[1, ] <- rep(1, popSize)
  pop[2, ] <- sample(tauclc, size = popSize)
  pop[3, ] <- N + 1

  return(pop)
}

#' The parents selection genetic algorithm operator for AMOC problem
#'
#' The genetic algorithm require to select a pair of chromosomes, representing
#' \code{dad} and \code{mom}, for the \code{crossover} operator to
#' produce offspring (individual for next generation). Here, the same linear
#' ranking method in \code{\link{selection_linear_rank}} is used to select a pair
#' of chromosomes for \code{dad} and \code{mom} in the at most one changepoint
#' (AMOC) problem. By default, the dad has better fit/smaller fitness function
#' value/larger rank than \code{mom}.
#' @param pop A matrix contains the chromosomes for all individuals. The number of
#' rows is equal to \code{lmax} and the number of columns is equal to the
#' \code{popSize}.
#' @param popFit A vector contains the objective function value (population fit)
#' being associated to each individual chromosome from above.
#' @return A list contains the chromosomes for \code{dad} and \code{mom}.
#' @import Rcpp
#' @import stats
#' @import graphics
#' @useDynLib changepointGA
#' @export
amoc_selection <- function(pop, popFit) {
  return(selection_linear_rank(pop, popFit))
}

#' Average crossover operator to produce offspring for AMOC problem
#'
#' In this crossover operator designed for AMOC problem, the new child is
#' produced by taking the average of the changepoint locations from \code{dad}
#' and \code{mom} and round to an integer. Note, every chromosome has at most
#' one candidate changepoint location.
#' @param mom Among two selected individuals, \code{mom} represents the selected
#' chromosome representation with lower fitness function value.
#' @param dad Among two selected individuals, \code{dad} represents the selected
#' chromosome representation with larger fitness function value.
#' @param prange The default value is \code{NULL}. If there is no requirement
#' on model order selection, such an auxiliary argument is needed for \code{GA}
#' and \code{IslandGA} functions.
#' @param minDist The minimum length between two adjacent changepoints.
#' @param lmax The maximum possible length of the chromosome representation.
#' @param N The length of time series.
#' @return The child chromosome that produced from \code{mom} and \code{dad} for
#' next generation.
#' @import Rcpp
#' @import stats
#' @import graphics
#' @useDynLib changepointGA
#' @export
amoc_crossover <- function(mom, dad, prange = NULL, minDist, lmax, N) {
  child <- matrix(0, nrow = lmax, 1)
  child[1] <- 1
  child[2] <- round((dad[2] + mom[2]) / 2)
  child[3] <- N + 1

  return(child)
}

#' Jump mutation operator to produce offspring for AMOC problem
#'
#' In a certain probability, the \code{mutation} genetic operator can be applied
#' to generate a new \code{child}. In this AMOC mutation operator, the new child
#' changepoint location can be down via a "jump" method. The child changepoint
#' location will jump \code{minDist} time units either to the left or right to
#' produce the changepoint location for the mutated child. The jump direction is
#' randomly decided with 0.5 probability.
#'
#' @param child The child chromosome resulting from the \code{crossover} genetic
#' operator.
#' @param prange The default value is \code{NULL}. If there is no requirement
#' on model order selection, such an auxiliary argument is needed for \code{GA}
#' and \code{IslandGA} functions.
#' @param minDist The minimum length between two adjacent changepoints in
#' \code{\link{amoc_selection}} operator, which is also the jump magnitude in the
#' \code{amoc_mutation} operator.
#' @param pchangepoint An auxiliary argument is needed for \code{GA}
#' and \code{IslandGA} functions.
#' @param lmax An auxiliary argument is needed for \code{GA} and \code{IslandGA}
#' functions.
#' @param mmax An auxiliary argument is needed for \code{GA} and \code{IslandGA}
#' functions.
#' @param N An auxiliary argument is needed for \code{GA} and \code{IslandGA}
#' functions.
#' @return The resulting child chromosome representation.
#' @import Rcpp
#' @import stats
#' @import graphics
#' @useDynLib changepointGA
#' @export
amoc_mutation <- function(child, prange = NULL, minDist, pchangepoint = NULL, lmax = NULL, mmax = NULL, N = NULL) {
  tmptau <- 1

  while (tmptau < floor(0.05 * N) | tmptau > ceiling(0.95 * N)) {
    tmpsign <- sample(x = c(-1, 1), size = 1)
    tmptau <- child[2] + tmpsign * minDist
  }
  child[2] <- tmptau

  return(child)
}

make_demo_data <- function(N = 80) {
  XMatT <- matrix(1, nrow = N, ncol = 1)
  colnames(XMatT) <- "intercept"
  
  Xt <- ts_sim(
    Ts = N,
    beta = 0.5,
    XMat = XMatT,
    sigma = 1,
    phi = 0.5,
    theta = NULL,
    Delta = c(2, -2),
    CpLoc = c(20, 60),
    seed = 1234
  )
  
  list(N = N, XMatT = XMatT, Xt = Xt)
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

arima_bic_changepointGA_rcpp <- function(chromosome_, XMat_, Xt_) {
    .Call(`_changepointGA_arima_bic_changepointGA_rcpp`, chromosome_, XMat_, Xt_)
}

rank_asR <- function(x, decreasing = FALSE) {
    .Call(`_changepointGA_rank_asR`, x, decreasing)
}

#' Randomly select the chromosome
#'
#' Randomly select the changepoint configuration for population initialization.
#' The selected changepoint configuration represents a changepoint chromosome.
#' The first element of the chromosome represent the number of changepoints
#' and the last non-zero element always equal to the length of time series + 1.
#'
#' @param N The length of time series.
#' @param prange A list object containing the possible range for other
#' pre-defined model parameters, i.e. AR/MA order of ARMA models.
#' @param minDist The minimum length between two adjacent changepoints.
#' @param pchangepoint Same as \code{Pchangepoint}, the probability that a changepoint has occurred.
#' @param mmax The maximum possible number of changepoints in the data set.
#' @param lmax The maximum possible length of the chromosome representation.
#' @return A single changepoint configuration format as above.
#' @export
select_tau <- function(N, prange, minDist, pchangepoint, mmax, lmax) {
    .Call(`_changepointGA_select_tau`, N, prange, minDist, pchangepoint, mmax, lmax)
}

#' Random population initialization
#'
#' Randomly generate the individuals' chromosomes (changepoint confirgurations)
#' to construct the first generation population.
#'
#' @param popSize An integer represents the number of individual in each
#' population for GA (or subpopulation for IslandGA).
#' @param prange Default is \code{NULL} for only changepoint detection. If
#' \code{prange} is specified as a list object, which contains the range of
#' each model order parameters for order selection (integers). The number of
#' order parameters must be equal to the length of \code{prange}.
#' @param N The length of time series.
#' @param minDist The minimum length between two adjacent changepoints.
#' @param pchangepoint Same as \code{pchangepoint} from \code{\link{cptga}} or 
#' \code{\link{cptgaisl}}, the probability that a changepoint has occurred.
#' @param mmax The maximum possible number of changepoints in the data set.
#' @param lmax The maximum possible length of the chromosome representation.
#' @details
#' Each population can be stored in a matrix with \code{lmax} rows and 
#' \code{popSize} columns, where each column represents an individual chromosome
#' in the format of
#' \deqn{C = (m, \boldsymbol{s}, \boldsymbol{\tau}, N+1)',} 
#' in \code{\link{cptga}} or \code{\link{cptgaisl}}. This function can randomly 
#' initialize the population matrix with some imposed constraints, such as 
#' ensuring the number of time points between two adjacent changepoints is 
#' greater than or equal to \code{minDist}. This prevents unrealistic scenario 
#' of two changepoints being too close to each other and helps reduce the total 
#' number of admissible solutions in the search space. Users can adjust the 
#' level of searching space reduction by setting an appropriate value for 
#' \code{minDist}. During population initialization, each changepoint location 
#' in \eqn{\boldsymbol{\tau}} can be selected sequentially. With a specified 
#' probability \code{pchangepoint} denoting the probability of a time point 
#' being selected as a changepoint, the first changepoint \eqn{\tau_{1}} is 
#' randomly picked between \eqn{t=1+minDist} and \eqn{t=N}. Then \eqn{\tau_{2}} 
#' is randomly selected from a smaller range between \eqn{t=\tau_{1}+minDist} 
#' and \eqn{t=N}. The process is continued until the last admissible changepoint 
#' \eqn{t=N-minDist} is exceeded, and the number of changepoints \eqn{m} is 
#' obtained automatically. For added flexibility, users can specify their own 
#' population initialization function. The default population initialization uses 
#' \code{\link{select_tau}} to select the chromosome for the first generation 
#' population.
#' @return A matrix that contains each individual's chromosome.
#' @export
random_population <- function(popSize, prange, N, minDist, pchangepoint, mmax, lmax) {
    .Call(`_changepointGA_random_population`, popSize, prange, N, minDist, pchangepoint, mmax, lmax)
}

#' Uniform crossover to produce offsprings
#'
#' In uniform crossover, typically, each bit is chosen from either parent with
#' equal probability. Other mixing ratios are sometimes used, resulting in
#' offspring which inherit more genetic information from one parent than the
#' other. In a uniform crossover, we don’t divide the chromosome into segments,
#' rather we treat each gene separately. In this, we essentially flip a coin
#' for each chromosome to decide whether or not it will be included in the
#' off-spring. If model order selection is requested, each child's model order
#' has the equal probability (0.5) from \code{dad} and \code{mom}.
#' @param mom Among two selected individuals, \code{mom} represents the selected
#' chromosome representation with lower fitness function value.
#' @param dad Among two selected individuals, \code{dad} represents the selected
#' chromosome representation with larger fitness function value.
#' @param prange Default value is \code{NULL} for only changepoint detection. If
#' \code{prange} is specified as a list object, which contains the range of
#' each model order parameters for order selection (integers). The number of
#' order parameters must be equal to the length of \code{prange}.
#' @param minDist The required minimum distance between two adjacent changepoints.
#' @param lmax The user specified maximum number of changepoints, by default,
#' as \code{N/2 - 1}.
#' @param N The length of time series.
#' @return The child chromosome that produced from \code{mom} and \code{dad} for
#' next generation.
#' @export
uniform_crossover <- function(mom, dad, prange, minDist, lmax, N) {
    .Call(`_changepointGA_uniform_crossover`, mom, dad, prange, minDist, lmax, N)
}

#' The default parents selection genetic algorithm operator
#'
#' The genetic algorithm require to select a pair of chromosomes, representing
#' \code{dad} and \code{mom}, for the \code{crossover} operator to
#' produce offspring (individual for next generation). The parents chromosomes
#' are randomly selectd from the initialized population by a linear ranking
#' method according to each individual's fittness in the input argument
#' \code{popFit}. By default, the dad has better fit/smaller fitness function
#' value/larger rank than \code{mom}.
#' @param pop A matrix contains the chromosomes for all individuals. The number of
#' rows is equal to \code{lmax} and the number of columns is equal to the
#' \code{popSize}.
#' @param popFit A vector contains the objective function value (population fit)
#' being associated to each individual chromosome from above.
#' @return A list contains the chromosomes for \code{dad} and \code{mom}.
#' @export
selection_linear_rank <- function(pop, popFit) {
    .Call(`_changepointGA_selection_linear_rank`, pop, popFit)
}

#' Example function: Calculating BIC for AR(1) model
#'
#' The objective function for changepoint search in Autoregressive
#' moving average with model order selection via Bayesian Information Criterion
#' (BIC).
#'
#' @param chromosome The chromosome consists of the number of changepoints, the
#' order of AR part, the order of MA part, the changepoint locations, and a
#' value of time series length plus 1 (N+1) indicating the end of the chromosome.
#' @param plen The number of model order parameters that need to be selected.
#' If model order selection needs to be performed simultaneously with the
#' changepoint detection task, \code{plen} should be nonzero.
#' @param XMat A matrix contains the covariates, but not includes changepoint
#' effects, for time series regression.
#' @param Xt The simulated ARMA time series from \code{ts.sim} function.
#' @return The BIC value of the objective function.
#' @import stats
#' @importFrom Rcpp sourceCpp
#' @useDynLib changepointGA
#' @export
#' @examples
#' Ts <- 1000
#' betaT <- c(0.5) # intercept
#' XMatT <- matrix(1, nrow = Ts, ncol = 1)
#' colnames(XMatT) <- "intercept"
#' sigmaT <- 1
#' phiT <- c(0.5)
#' thetaT <- NULL
#' DeltaT <- c(2, -2)
#' Cp.prop <- c(1 / 4, 3 / 4)
#' CpLocT <- floor(Ts * Cp.prop)
#'
#' myts <- ts_sim(
#'   Ts = Ts, beta = betaT, XMat = XMatT, sigma = sigmaT, phi = phiT, theta = thetaT,
#'   Delta = DeltaT, CpLoc = CpLocT, seed = 1234
#' )
#'
#' # candidate changepoint configuration
#' chromosome <- c(2, 250, 750, 1001)
#' arima_bic(chromosome, XMat = XMatT, Xt = myts)
arima_bic <- function(chromosome, plen = 0, XMat, Xt) {
  arima_bic_changepointGA_rcpp(chromosome, XMat, Xt)
}

#' Time series simulation with changepoint effects
#'
#' This is a function to simulate time series with changepoint effects. See
#' details below.
#' 
#' @param Ts A integer indicating simulated time series length (sample size).
#' @param beta A parameter vector contains other mean function parameters without changepoint parameters.
#' @param XMat The covairates for time series mean function without changepoint indicators.
#' @param sigma The standard deviation for time series residuals \eqn{\epsilon_{t}}.
#' @param phi A vector for the autoregressive (AR) parameters for AR part.
#' @param theta A vector for the moving average (MA) parameters for MA part.
#' @param d A nonnegative integer for the order of differencing in the ARIMA model. Default is 0, which reduces to an ARMA model.
#' @param Delta The parameter vector contains the changepoint parameters for time series mean function.
#' @param CpLoc A vector contains the changepoint locations range from \eqn{1\leq\tau\leq T_{s}}.
#' @param seed The random seed for simulation reproducibility.
#' @details
#' The simulated time series \eqn{Z_t, t = 1, \ldots, T_s} is generated from
#' \deqn{Z_t = \mu_t + e_t.}
#'
#' The error process \eqn{e_t} can be:
#' \itemize{
#'   \item IID Gaussian noise when \code{phi = NULL}, \code{theta = NULL},
#'     and \code{d = 0}.
#'   \item An ARMA(p,q) process when \code{d = 0} and at least one of
#'     \code{phi} or \code{theta} is specified.
#'   \item An ARIMA(p,d,q) process when \code{d > 0}.
#' }
#'
#' The mean structure \eqn{\mu_t} is specified through \code{XMat} and
#' \code{beta}, and changepoint effects can be introduced as
#' \deqn{\mu_t = X_t^\top \beta + \Delta_1 I(t \ge \tau_1) + \cdots +
#' \Delta_m I(t \ge \tau_m),}
#' where \eqn{1 \le \tau_1 < \cdots < \tau_m \le T_s} are changepoint
#' locations and \eqn{\Delta_1, \ldots, \Delta_m} are changepoint magnitudes.
#'
#' @return The simulated time series with attributes:
#' \itemize{
#'   \item \code{DesignX}: The covariates including changepoint indicators.
#'   \item \code{mu}: The mean vector for the simulated time series.
#'   \item \code{theta}: The true parameter vector including changepoint effects.
#'   \item \code{CpEff}: The true changepoint magnitudes.
#'   \item \code{CpLoc}: The true changepoint locations.
#'   \item \code{arEff}: The AR coefficients.
#'   \item \code{maEff}: The MA coefficients.
#'   \item \code{diffEff}: The differencing order.
#'   \item \code{seed}: The random seed used for simulation.
#' }
#'
#' @import Rcpp
#' @import stats
#' @useDynLib changepointGA
#' @export
#'
#' @examples
#' ##### M1: Time series observations are IID
#' Ts <- 1000
#' betaT <- c(0.5)
#' XMatT <- matrix(1, nrow = Ts, ncol = 1)
#' colnames(XMatT) <- "intercept"
#' sigmaT <- 1
#' DeltaT <- c(2, -2)
#' Cpprop <- c(1 / 4, 3 / 4)
#' CpLocT <- floor(Ts * Cpprop)
#'
#' myts <- ts_sim(
#'   Ts = Ts, beta = betaT, XMat = XMatT, sigma = sigmaT,
#'   Delta = DeltaT, CpLoc = CpLocT, seed = 1234
#' )
#'
#' ##### M2: ARMA(2,1) model with constant mean
#' Ts <- 1000
#' betaT <- c(0.5)
#' XMatT <- matrix(1, nrow = Ts, ncol = 1)
#' colnames(XMatT) <- "intercept"
#' sigmaT <- 1
#' phiT <- c(0.5, -0.5)
#' thetaT <- c(0.8)
#' DeltaT <- c(2, -2)
#' CpLocT <- floor(Ts * c(1 / 4, 3 / 4))
#'
#' myts <- ts_sim(
#'   Ts = Ts, beta = betaT, XMat = XMatT, sigma = sigmaT,
#'   phi = phiT, theta = thetaT, d = 0,
#'   Delta = DeltaT, CpLoc = CpLocT, seed = 1234
#' )
#'
#' ##### M3: ARIMA(1,1,1) model with changepoint effects
#' Ts <- 1000
#' betaT <- c(0.5)
#' XMatT <- matrix(1, nrow = Ts, ncol = 1)
#' colnames(XMatT) <- "intercept"
#' sigmaT <- 1
#' phiT <- c(0.5)
#' thetaT <- c(-0.3)
#' DeltaT <- c(2, -2)
#' CpLocT <- floor(Ts * c(1 / 4, 3 / 4))
#'
#' myts <- ts_sim(
#'   Ts = Ts, beta = betaT, XMat = XMatT, sigma = sigmaT,
#'   phi = phiT, theta = thetaT, d = 1,
#'   Delta = DeltaT, CpLoc = CpLocT, seed = 1234
#' )
ts_sim <- function(Ts, beta, XMat, sigma, phi = NULL, theta = NULL, d = 0,
                   Delta = NULL, CpLoc = NULL, seed = NULL) {
  if (!is.null(seed)) {
    set.seed(seed)
  }
  
  if (is.null(Delta)) {
    if (is.null(CpLoc)) {
      warning("\n No changepoint effects!\n")
      DesignX <- XMat
      mu <- DesignX %*% beta
    } else {
      stop("Changepoint effects invalid!")
    }
  } else {
    if (is.null(CpLoc)) {
      stop("Changepoint effects invalid!")
    } else {
      if (any(CpLoc > Ts) | any(CpLoc < 0)) {
        stop("Changepoint effects invalid!")
      }
      if (length(Delta) != length(CpLoc)) {
        stop("Changepoint effects invalid!")
      }
      # changepoints
      tmptau <- unique(c(CpLoc, Ts + 1))
      CpMat <- matrix(0, nrow = Ts, ncol = length(tmptau) - 1)
      for (i in 1:NCOL(CpMat)) {
        CpMat[tmptau[i]:(tmptau[i + 1] - 1), i] <- 1
      }
      DesignX <- cbind(XMat, CpMat)
      beta <- c(beta, Delta)
      mu <- DesignX %*% beta
    }
  }
  
  if (is.null(phi) & is.null(theta) & d == 0) {
    et <- rnorm(n = Ts, mean = 0, sd = sigma) # independent
  } else {
    p <- if (is.null(phi)) 0 else length(phi)
    q <- if (is.null(theta)) 0 else length(theta)
    
    et <- arima.sim(
      n = Ts,
      model = list(order = c(p, d, q), ar = phi, ma = theta),
      sd = sigma
    )
    
    et <- as.numeric(et)
    if (length(et) > Ts) {
      et <- tail(et, Ts)
    }
  }
  
  Z <- et + mu
  
  attr(Z, "DesignX") <- DesignX
  attr(Z, "mu") <- mu
  attr(Z, "theta") <- beta
  attr(Z, "CpEff") <- Delta
  attr(Z, "CpLoc") <- CpLoc
  attr(Z, "arEff") <- phi
  attr(Z, "maEff") <- theta
  attr(Z, "diffEff") <- d
  attr(Z, "seed") <- seed
  
  return(Z)
}

test_that("ts_sim is reproducible with the same seed", {
  Ts <- 100
  XMatT <- matrix(1, nrow = Ts, ncol = 1)
  colnames(XMatT) <- "intercept"
  
  sim1 <- ts_sim(
    Ts = Ts,
    beta = 0.5,
    XMat = XMatT,
    sigma = 1,
    Delta = c(2, -2),
    CpLoc = c(25, 75),
    seed = 1234
  )
  
  sim2 <- ts_sim(
    Ts = Ts,
    beta = 0.5,
    XMat = XMatT,
    sigma = 1,
    Delta = c(2, -2),
    CpLoc = c(25, 75),
    seed = 1234
  )
  
  expect_equal(sim1, sim2)
})

test_that("ts_sim returns numeric output with expected length", {
  Ts <- 100
  XMatT <- matrix(1, nrow = Ts, ncol = 1)
  colnames(XMatT) <- "intercept"
  
  sim <- ts_sim(
    Ts = Ts,
    beta = 0.5,
    XMat = XMatT,
    sigma = 1,
    Delta = c(2, -2),
    CpLoc = c(25, 75),
    seed = 1234
  )
  
  expect_true(is.numeric(sim))
  expect_length(sim, Ts)
})

test_that("ts_sim works with ARMA terms", {
  Ts <- 100
  XMatT <- matrix(1, nrow = Ts, ncol = 1)
  colnames(XMatT) <- "intercept"
  
  sim <- ts_sim(
    Ts = Ts,
    beta = 0.5,
    XMat = XMatT,
    sigma = 1,
    phi = c(0.5, -0.5),
    theta = 0.8,
    d = 0,
    Delta = c(2, -2),
    CpLoc = c(25, 75),
    seed = 1234
  )
  
  expect_true(is.numeric(sim))
  expect_length(sim, Ts)
  expect_equal(attr(sim, "diffEff"), 0)
})

test_that("ts_sim works with ARIMA(1,1,1) terms", {
  Ts <- 100
  XMatT <- matrix(1, nrow = Ts, ncol = 1)
  colnames(XMatT) <- "intercept"
  
  sim <- ts_sim(
    Ts = Ts,
    beta = 0.5,
    XMat = XMatT,
    sigma = 1,
    phi = 0.5,
    theta = -0.3,
    d = 1,
    Delta = c(2, -2),
    CpLoc = c(25, 75),
    seed = 1234
  )
  
  expect_true(is.numeric(sim))
  expect_length(sim, Ts)
  expect_equal(attr(sim, "arEff"), 0.5)
  expect_equal(attr(sim, "maEff"), -0.3)
  expect_equal(attr(sim, "diffEff"), 1)
})

---
title: "A brief guide to changepointGA"
author: "Mo Li"
date: "`r format(Sys.time(), '%d %b %Y')`"
output: 
  rmarkdown::html_vignette:
    number_sections: false
    toc: true
vignette: >
  %\VignetteIndexEntry{A brief guide to changepointGA}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Introduction

The **changepointGA** package applies Genetic Algorithms (GA) to detect both single and multiple changepoints in time series data. Each candidate solution encodes changepoint locations as integer-valued time indices, allowing for flexible representation of changepoint configurations. Because each changepoint introduces at least one associated model parameter, the total number of parameters is not fixed in the multiple changepoint setting. Consequently, the model parameter vector must adapt dynamically as the number and location of changepoints vary. Unlike fixed-length encodings, **changepointGA** supports variable-length chromosome representations. By encoding each changepoint as an integer, the algorithm dynamically adjusts the dimensionality of the model parameter vector and jointly optimizes the model fitness function. Additionally, in many classical time series models—such as autoregressive (AR), moving average (MA), autoregressive moving average (ARMA), and periodic autoregression (PAR)—model selection traditionally involves determining the appropriate order (e.g., the AR or MA lag length) before detecting changepoints. This separate step can introduce additional complexity and potential error. In contrast, our **changepointGA** can simultaneously estimates integer-valued model hyperparameters (e.g., model orders), model parameters, and changepoint configurations. This integrated approach eliminates the need for a separate model selection phase and supports a more unified and efficient framework for changepoint detection.

```{r}
library(changepointGA)

set.seed(1234)
```

## Simulate a time series data with changepoints

This package includes the time series simulation function `ts_sim()` with the ability to introduce changepoint effects. The time series \eqn{Z_{t}, t=1,\ldots,T_{s}} could be simulated from a class of time series models,

$$Z_{t}=\mu_{t}+e_{t}.$$

The model could be specified by the arguments of `phi` and `theta`, representing the AR coefficients of and the MA coefficients for the ARMA model. The default values of `phi` and `theta` are `NULL`, generating time series with $e_{t}$ that are independent and identically $N(0,\sigma^{2})$ distributed. If we assign values to `phi` and/or `theta`, the $e_{t}$ will be generated from an ARMA($p$,$q$) process, where $p$ equals the number of values in `phi` and $q$ equals the number of values in `theta`. For the above model, $\mu_{t}$ denotes time-varying mean function and the mean changepoint effects could be introduced through indicator functions to $\mu_{t}$. Note that the **changepointGA** could also work with other time series model as long as the fitness function is appropriately specified. To illustrate the package usage, we will generate the $e_{t}$ following an ARMA($p=1$,$q=1$) process. The AR coefficient is $\phi=0.5$ and the MA coefficient is $\theta=0.8$. The time varying mean values are generated through the following equation,

$$\mu_{t}=    \begin{cases}
        \beta_0, 
        & 1 \leq t \leq \tau_1, \\
        \beta_0 + \Delta_1, 
        & \tau_1 + 1 \leq t \leq \tau_2, \\
        \beta_0 + \Delta_2, 
        & \tau_2 + 1 \leq t \leq N.
    \end{cases}$$

including the intercept term with $\beta_{0}=0.5$ and two changepoints at $\tau_{1}=125$ and $\tau_{2}=375$ with the mean changing effects of $\Delta_{1}=2$ and $\Delta_{2}=-2$. The users can also include additional covariates into matrix (`XMatT` below) for other possible model dynamics, such as tread and seasonalities.

```{r}
Ts <- 200
betaT <- c(0.5) # intercept
XMatT <- matrix(rep(1, Ts), ncol = 1)
colnames(XMatT) <- c("intercept")
sigmaT <- 1
phiT <- c(0.5)
thetaT <- c(0.8)
DeltaT <- c(2, -2)
CpLocT <- c(50, 150)

myts <- ts_sim(Ts=Ts, beta = betaT, XMat = XMatT, sigma = sigmaT, phi = phiT, theta = thetaT, Delta = DeltaT, CpLoc = CpLocT, seed = 1234)
str(myts)
```

The simulated time series could be plotted in the figure below. The vertical dashed blue lines indicate the true changepoint location and the horizontal red dashed segments represent the time series segment sample mean. We can clearly observe the mean changing effects from the introduced changepoints $\tau_{1}=250$ and $\tau_{2}=500$. The **changepointGA** package will be used to detected these two changepoints.

```{r, fig.align = "center", fig.height=4, fig.width=6}
plot(x = 1:Ts, y = myts, type = "l", xlab = "Time", ylab = "Y")
abline(v = CpLocT, lty = "dashed", col = "blue", lwd = 2)

## Segmentation sample mean calculation and plotting
m <- length(CpLocT)
tauclc <- c(1, CpLocT, Ts + 1)
seg.len <- diff(tauclc)
ff <- rep(0:m, times = seg.len)
Xseg <- split(myts, ff)
mu.seg <- unlist(lapply(Xseg, mean), use.names = F)
for (i in 1:(m + 1)) {
  segments(
    x0 = tauclc[i], y0 = mu.seg[i],
    x1 = tauclc[i + 1], y1 = mu.seg[i],
    col = "red", lty = "dashed", lwd = 2
  )
}
```

## The objective function for optimization

Considering the model used in the data simulation, we could use the R function, `ARIMA.BIC.Order()`, included in the **changepointGA** package. This function inputs include the 

- `chromosome`, consists of the number of changepoints, the order of AR component (refers to the number of lagged terms used to model the current value of a time series), the order of MA component (refers to the number of lagged error terms used to model the current value of a time series), the changepoint locations, and a value of time series sample size plus 1 ($N+1$) indicating the end of the chromosome;
- `plen` represents the number of model order hyperparameters. We have `plen=2` for this example representing one hyperparameter for AR order and one hyperparameter for MA order;
- `XMat` requires the design matrix, including the of-interested covariates;
- `Xt` requires the simulated time series data objects.

Given a specific chromosome that encodes the ARMA model structure—namely, the AR and moving average (MA) orders—as well as the changepoint configuration, the function ARIMA.BIC.Order() evaluates the BIC for the corresponding segmented model. The returned BIC value serves as the fitness score of that configuration. The example below presents the BIC value computed at the true changepoint locations with the true ARMA model orders, providing a benchmark for comparison.

```{r}
arima_bic_order_pq(chromosome = c(2, 1, 1, 50, 150, Ts + 1), plen = 2, XMat = XMatT, Xt = myts)
```

## Basic GA

The GA requires users to set up an objective function (`ObjFunc`), sample size of the time series (`N`). All other algorithm hyperparameters, such as `popSize`, `pcrossover`, `pmutation` and etc, can be specified through the function arguments. Please use `?cptga` to see help detailed documentation. 

A special note is warranted regarding the setup of the `prange` argument:

- For tasks requiring changepoint detection only, the default value is `NULL`.
- If joint estimation of model order and changepoints is required, `prange` must be specified as a list, where each element defines the permissible range for one model order parameter. The length of this list must match the number of model order parameters to be estimated in the objective function.

As shown below, we will have 40 individual chromosomes in the population for each generation. `prange` is a list object containing integer values ranges for parameters `phi` and `theta`. For example, in the code snippet below, `prange` specifies that the AR and MA orders can vary from the set $\{0, 1, 2\}$. 

```{r}
N <- Ts
prange <- list(ar = c(0, 2), ma = c(0, 2))
```

### Population chromosome initialization

The population initialization step generates the initial set of chromosomes for the GA. This is controlled by the population operator by the argument `popInitialize`, which is one of the four required components in the GA or island model GA for the `cptga()` function—namely: population, selection, crossover, and mutation.

By default, the population operator randomly generates a matrix-like R object, where each column represents a chromosome (i.e., a candidate solution). The matrix should have `lmax` rows and `popSize` columns, where `lmax` is the maximum chromosome length and `popSize` is the number of individuals in the population.

To improve performance or guide the search, users may replace the default population operator by supplying a custom initialization matrix with the same dimensions. Providing informed or domain-specific initial solutions can lead to faster convergence and improved optimization outcomes.

- `suggestions` is a list object where each element specifies a vector of changepoint locations for one chromosome.
- The default value is `NULL`, meaning the population is entirely randomly initialized.
- If `length(suggestions) == popSize`, all chromosomes in the population will be taken from suggestions.
- If `length(suggestions) < popSize`, the remaining chromosomes will be filled in using the default popInitialize method.
- The number of provided suggestions cannot exceed popSize.

Providing better or more informed suggestions can help the Genetic Algorithm converge faster by giving it a more promising starting point.

```{r}
suggestions <- list(NULL, c(50), c(50, 150), c(50, 100, 150))
```

Each element of the list corresponds to a candidate changepoint configuration. In this example, 4 suggestions are provided:

- `NULL` indicates no changepoint;
- `c(50)` suggests a single changepoint at time 50;
- `c(50, 150)` specifies changepoints at time 50 and 150, and etc,

and 6 chromosomes will be generated randomly using the default function if `popSize=10`. These suggested chromosomes can be incorporated into the population by replacing the default population operator or modifying the population matrix directly.

Since we are trying to estimate the changepoint configurations, it is necessary to set the `XMat` to include all other covariates except the changepoint indicators, as if we are under a changepoint free model.

```{r}
XMatEst <- matrix(1, nrow = N, ncol = 1)
```

Then we call function `cptga()` for changepoint searching and simultaneously estimate the best fitted model AR and MA orders. The estimated number of changepoints and corresponding changepoint locations could be extracted as below.

```{r}
reschangepointGA <- suppressWarnings(cptga(
  ObjFunc = arima_bic_order_pq,
  N = N,
  prange = prange,
  suggestions = suggestions,
  option = "both",
  XMat = XMatEst,
  Xt = myts
))
print(reschangepointGA)
summary(reschangepointGA)
```


## Island model GA

The island model GA can be implemented using the `cptgaisl()` function, which supports similar core arguments as the basic GA function `cptga()`, but introduces additional flexibility and parallelism via multiple isolated sub-populations (islands).

Key differences include:

- The initial population in `cptgaisl()` is expected to be a list object, where each element is a matrix and corresponds to the population for one island.
- The length of this list must equal the number of islands (e.g., `numIslands`).
- By default, each island's population is initialized using the internal operator, but users may insert their own informed or "reasonable" initial solutions into the `suggestions` list as above. These will serve as seed chromosomes for each island's initial population.

If model order selection is enabled (e.g., for AR, MA, or ARIMA models), the associated model orders will be randomly sampled based on the user-supplied `prange` argument.

```{r}
tim1 <- Sys.time()
resIslandchangepointGA <- suppressWarnings(cptgaisl(
  ObjFunc = arima_bic_order_pq,
  N = N,
  prange = prange,
  popSize = 160,
  numIslands = 2,
  maxMig = 1000,
  maxgen = 50,
  maxconv = 20,
  option = "both",
  XMat = XMatEst,
  Xt = myts
))
tim2 <- Sys.time()
tim2 - tim1
print(resIslandchangepointGA)
summary(resIslandchangepointGA)
plot(resIslandchangepointGA, data = myts)
```

## Parallel computing

The default behavior of both `cptga()` and `cptgaisl()` is to evaluate each candidate changepoint configuration sequentially, which can be computationally expensive, especially during the initial population fitness evaluation. To address this, users may enable parallel computing to significantly speed up the evaluation process.

Parallel execution requires the R packages `foreach` and `doParallel`. This approach is particularly beneficial when the population size and the number of islands are large, as more evaluations can be distributed across available CPU cores.

To enable parallel computing, simply set the argument `parallel = TRUE` and specify the number of cores via `nCore`. It is generally recommended to set `nCore` equal to the number of islands (when using `cptgaisl()`), or to the number of physical cores available for `cptga()`. The following example will distribute the fitness evaluations across two parallel threads, potentially reducing the runtime substantially for large or complex problems.

```{r}
tim3 <- Sys.time()
resIslandchangepointGA <- suppressWarnings(cptgaisl(
  ObjFunc = arima_bic_order_pq,
  N = N,
  prange = prange,
  popSize = 160,
  numIslands = 2,
  maxMig = 1000,
  maxgen = 50,
  maxconv = 20,
  option = "both",
  parallel = TRUE,
  nCore = 2,
  XMat = XMatEst,
  Xt = myts
))

tim4 <- Sys.time()
tim4 - tim3
print(resIslandchangepointGA)
summary(resIslandchangepointGA)
plot(resIslandchangepointGA, data = myts)
```

## Changepoint configuration distance calculation

It is important to measure how different the estimated changepoint configuration away from the true configuration used in data generation. The pairwise distance metric proposed by Shi et al. (2022) can help quantify such differences. We implemented and included the method via the `cpDist()` function in this package.

```{r}
truetau <- c(50, 150)
esttau <- c(45, 147)
cpt_dist(tau1 = truetau, tau2 = esttau, N = N)
```


## References

Shi, X., Gallagher, C., Lund, R., & Killick, R. (2022). A comparison of single and multiple changepoint techniques for time series data. *Computational Statistics & Data Analysis*, 170, 107433.

```{r}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/arima_bic.R
\name{arima_bic}
\alias{arima_bic}
\title{Example function: Calculating BIC for AR(1) model}
\usage{
arima_bic(chromosome, plen = 0, XMat, Xt)
}
\arguments{
\item{chromosome}{The chromosome consists of the number of changepoints, the
order of AR part, the order of MA part, the changepoint locations, and a
value of time series length plus 1 (N+1) indicating the end of the chromosome.}

\item{plen}{The number of model order parameters that need to be selected.
If model order selection needs to be performed simultaneously with the
changepoint detection task, \code{plen} should be nonzero.}

\item{XMat}{A matrix contains the covariates, but not includes changepoint
effects, for time series regression.}

\item{Xt}{The simulated ARMA time series from \code{ts.sim} function.}
}
\value{
The BIC value of the objective function.
}
\description{
The objective function for changepoint search in Autoregressive
moving average with model order selection via Bayesian Information Criterion
(BIC).
}
\examples{
Ts <- 1000
betaT <- c(0.5) # intercept
XMatT <- matrix(1, nrow = Ts, ncol = 1)
colnames(XMatT) <- "intercept"
sigmaT <- 1
phiT <- c(0.5)
thetaT <- NULL
DeltaT <- c(2, -2)
Cp.prop <- c(1 / 4, 3 / 4)
CpLocT <- floor(Ts * Cp.prop)

myts <- ts_sim(
  Ts = Ts, beta = betaT, XMat = XMatT, sigma = sigmaT, phi = phiT, theta = thetaT,
  Delta = DeltaT, CpLoc = CpLocT, seed = 1234
)

# candidate changepoint configuration
chromosome <- c(2, 250, 750, 1001)
arima_bic(chromosome, XMat = XMatT, Xt = myts)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/amoc_func.R
\name{amoc_population}
\alias{amoc_population}
\title{Random population initialization for AMOC problem}
\usage{
amoc_population(popSize, prange, N, minDist, pchangepoint, mmax, lmax)
}
\arguments{
\item{popSize}{An integer represents the number of individual in each
population for GA (or subpopulation for IslandGA).}

\item{prange}{Default is \code{NULL} for only changepoint detection. If
\code{prange} is specified as a list object, which contains the range of
each model order parameters for order selection (integers). The number of
order parameters must be equal to the length of \code{prange}.}

\item{N}{The length of time series.}

\item{minDist}{The minimum length between two adjacent changepoints.}

\item{pchangepoint}{The probability that a changepoint can occur.}

\item{mmax}{The maximum possible number of changepoints in the data set.}

\item{lmax}{The maximum possible length of the chromosome representation.}
}
\value{
A matrix that contains each individual's chromosome.
}
\description{
Randomly generate the individuals' chromosomes (changepoint configurations)
to construct the first generation population for the at most one changepoint
(AMOC) problem.
}
\details{
Given the possible candidate changepoint location set, each chromosome in the
first generation population can be obtained by randomly sampling one location
from the candidate set. The first element of every chromosome represent the
number of changepoints and the last non-zero element always equal to the
length of time series plus one (N+1).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{selection_linear_rank}
\alias{selection_linear_rank}
\title{The default parents selection genetic algorithm operator}
\usage{
selection_linear_rank(pop, popFit)
}
\arguments{
\item{pop}{A matrix contains the chromosomes for all individuals. The number of
rows is equal to \code{lmax} and the number of columns is equal to the
\code{popSize}.}

\item{popFit}{A vector contains the objective function value (population fit)
being associated to each individual chromosome from above.}
}
\value{
A list contains the chromosomes for \code{dad} and \code{mom}.
}
\description{
The genetic algorithm require to select a pair of chromosomes, representing
\code{dad} and \code{mom}, for the \code{crossover} operator to
produce offspring (individual for next generation). The parents chromosomes
are randomly selectd from the initialized population by a linear ranking
method according to each individual's fittness in the input argument
\code{popFit}. By default, the dad has better fit/smaller fitness function
value/larger rank than \code{mom}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cptgaisl.R
\name{cptgaisl}
\alias{cptgaisl}
\title{Island model based genetic algorithm}
\usage{
cptgaisl(
  ObjFunc,
  N,
  prange = NULL,
  popSize = 200,
  numIslands = 5,
  pcrossover = 0.95,
  pmutation = 0.3,
  pchangepoint = 0.01,
  minDist = 1,
  mmax = NULL,
  lmax = NULL,
  maxMig = 1000,
  maxgen = 50,
  maxconv = 100,
  option = "cp",
  monitoring = FALSE,
  parallel = FALSE,
  nCore = NULL,
  tol = 1e-05,
  seed = NULL,
  popInitialize = "random_population",
  suggestions = NULL,
  selection = "selection_linear_rank",
  crossover = "uniform_crossover",
  mutation = "mutation",
  ...
)
}
\arguments{
\item{ObjFunc}{The fitness function to be minimized. Users can specify any R or Rcpp
function as the fitness function, setting the input as the potential solution to
the optimization problem and returning a numerical value as the output/fitness.
Depending on the user-specified chromosome representation, the optimization task
can be changepoint detection only or changepoint detection plus model order selection,
which can be specified via the \code{option} parameter. When
\code{option="both"}, the list \code{prange} must be specified to give the range
of model orders.}

\item{N}{The sample size of the time series.}

\item{prange}{The default value is \code{NULL} for changepoint detection only task.
If both model order selection and changepoint detection are required, the \code{prange}
argument should be provided as a list. Each element in this list must specify the range
for a corresponding model order parameter, and the length of the list object of
\code{prange} must match the number of order parameters to be estimated.}

\item{popSize}{An integer representing the total number of individuals in each generation, which
equal to the number of islands multiplied by the size of each island (i.e.,
\code{popSize = numIslands × Islandsize}).}

\item{numIslands}{An integer representing the number of islands (sub-populations).}

\item{pcrossover}{The probability that the crossover operator will apply to
the two selected parents' chromosomes to produce the offspring. The typical
value is close to 1, with the default setting in this package being 0.95.}

\item{pmutation}{The probability that the mutation operator applies on one
individual chromosome. Similar to the natural mutation process, new genetic
information is introduced to the offspring chromosome with a relatively small
probability (close to 0), with a default value of 0.3.}

\item{pchangepoint}{The probability that a changepoint has occurred. User
could change this probability based on domain knowledge and the time series
length. This probability is used during population initialization and in the
creation of new chromosomes by the mutation operator. By default, the mutation
operator function generates a new individual as the mutated offspring.}

\item{minDist}{The minimum length between two adjacent changepoints. Default value equals to one.}

\item{mmax}{The maximum number of changepoints allowed in the time series data
corresponds to the maximum possible length of \eqn{\boldsymbol{\tau}}.
For a time series of length 1000 and we only want to detect the changepoint
(\code{option="cp"}), the default value is 499. The suggested value should be
based on the length of the time series. For instance, if a time series has
length \eqn{N}, the recommended \code{mmax} should be \eqn{N/2-1}. It is suggested to
add the number of model hyperparameters if both changepoint detection and
model order selection tasks are of-interested simultaneously
(\code{option="both"}).}

\item{lmax}{The maximum possible length of the chromosome representation.
For a time series of length 1000 and we only want to detect the changepoint
(\code{option="cp"}), the default value is 501. The suggested value should be
based on the length of the time series. For instance, if a time series has
length \eqn{N}, the recommended \code{lmax} should be \eqn{2+N/2-1}. It is suggested to
add the number of model hyperparameters if both changepoint detection and
model order selection tasks are of-interested simultaneously (\code{option="both"}).}

\item{maxMig}{An integer indicates the maximum number of migrations. After
conducting \code{maxMig} migrations, the island model GA stops.}

\item{maxgen}{An integer indicates the maximum number of generations that each
island (subpopulation) undergoes before migration. It also determines the
requency of migration. The migration will occur after \code{maxgen} generations
for each island (sub-population).}

\item{maxconv}{An integer value is also used for algorithm termination. If
the overall best-fitted value remains unchanged after \code{maxconv}
consecutive migrations, the island model GA will terminate.}

\item{option}{A character string controls the optimization task. \code{"cp"}
indicates the task is changepoint detection only. \code{"both"} indicates the
task will include both changepoint detection and model order selection.}

\item{monitoring}{A logical value with either \code{TRUE} or \code{FALSE},
indicating whether to print out summarized results (current best fitness
function value and its corresponding $C$) for each generation from the GA.}

\item{parallel}{A logical value with \code{TRUE} or \code{FALSE}, indicating
whether to use multiple cores for parallel computation. Different from the
basic GA, in this approach, each island evolves independently. New population
generation—including parent selection, crossover, and mutation—is processed
independently and in parallel for each island, further saving computation
time. Once the new population generation is complete, the migration operator
is performed sequentially.}

\item{nCore}{An integer indicates the number of cores utilized in parallel
computing. For the island model GA, the number of cores used in parallel is
set to the \code{numIslands} for convenience.}

\item{tol}{An numerical value with the default value of \code{1e-05}. The
tolerance level that helps evaluate whether the two iterations have the same
fitness value, which aids in determining GA termination.}

\item{seed}{An integer with the default value of \code{NULL}. An single
integer allows function produce reproducible results.}

\item{popInitialize}{A function or sourced function name character string.
It should be designed for initializing a population. The default population
initialization is random initialization with some imposed constraints. See
\code{\link{random_population}} for example. The function returned object is
a matrix, \code{pop}. The users can specified their own \code{population}
function. It could also be a matrix object, which contain the user specified
chromosome. By default, each column represents one individual chromosome.
See \code{\link{random_population}} for details.}

\item{suggestions}{A list object. Default value is \code{NULL}. Each element
includes better or more reasonable guessed changepoints locations, which will
resulting in one chromosome. If the number of provided suggestions equals to
\code{popSize}, all the suggestions will be used as population. If the number
of provided suggestions is less than \code{popSize}, the function from
\code{popInitialize} will generate the remaining individual chromosomes.
The number of provided suggestions cannot be greater than \code{popSize}.
Having better \code{suggestions} can help GA converges faster.}

\item{selection}{A function or sourced function name character string. This
GA operator can help select \code{mom} and \code{dad} from current generation
population, where \code{dad} is set to have better fit (smaller fitness
function values). The default for selection uses the linear rank selection
method. See \code{\link{selection_linear_rank}} for example. The function
returned object is a list contain the chromosomes for \code{mom} and
\code{dad}.}

\item{crossover}{A function or sourced function name character string. This
GA operator can apply crossover to the chosen parents to produce child for
next generation with specified probability. The default for crossover uses
the uniform crossover method. See \code{\link{uniform_crossover}} for details
in the default crossover operator. The function returned object is a vector
contain the chromosomes for \code{child}.}

\item{mutation}{A function or sourced function name character string. This
GA operator can apply mutation to the produced \code{child} with the
specified probability \code{pmutation}. See \code{\link{mutation}} for
details in the default mutation operator. The function returned object
is a vector contain \code{child} chromosome representation.}

\item{...}{additional arguments that will be passed to the fitness function.}
}
\value{
Return an object class \code{cptgaisl-class}. See
\code{\link{cptgaisl-class}} for a more detailed description.
}
\description{
Perform the modified island-based genetic algorithm (IslandGA) for multiple changepoint detection.
Minimization of an objective function using genetic algorithm (GA).
The algorithm can be run sequentially or in explicit parallelisation.
}
\details{
For any pre-specified time series model with a specified set of changepoint locations,
model fit is evaluated using a fitness function \eqn{Q(\boldsymbol{\theta})},
where \eqn{\boldsymbol{\theta}=(\boldsymbol{s},\boldsymbol{\tau},\boldsymbol{\beta})'}
denotes the full parameter vector. Here, \eqn{\boldsymbol{s}} denotes the set
of model hyperparameters, potentially including the AR or MA orders, the degree
of ARIMA differencing, or the periodic AR order for PAR models. The vector
\eqn{\boldsymbol{\tau}=\{\tau_{1}, \ldots, \tau_{m}\}} specifies the
changepoint locations, with the number of changepoints \eqn{m} inferred as
part of the estimation. Each individual chromosome representation is specified as a vector,
\deqn{C = (m, \boldsymbol{s}, \boldsymbol{\tau}, N+1)',}
where \eqn{m} represents the number of changepoints and is also equivalent to
the length of vector \eqn{\boldsymbol{\tau}} containing all the changepoint
locations. \eqn{\boldsymbol{s}} contains the integer-valued parameter for time
series model structure specification, such as AR, MA, or PAR orders.
If no model order selection is desired, then \eqn{\boldsymbol{s}} is omitted
and the GA detects changepoints only. The changepoint locations in
\eqn{\boldsymbol{\tau}} are encoded as interger values between 2 and \eqn{N},
allowing the length of \eqn{\boldsymbol{\tau}} to vary dynamically with \eqn{m}.
The value \eqn{N+1} is appended at the end of \eqn{C} to serve as a delimiter
marking the boundary of the chromosome.
}
\examples{
\donttest{

Ts <- 1000
XMatT <- matrix(1, nrow = Ts, ncol = 1)
Xt <- ts_sim(
  Ts = Ts, beta = 0.5, XMat = XMatT, sigma = 1, phi = 0.5, theta = NULL,
  Delta = c(2, -2), CpLoc = c(250, 750), seed = 1234
)

## Multiple changepoint detection without model order selection

# without suggestions
GAISLres <- cptgaisl(ObjFunc = arima_bic, N = Ts, XMat = XMatT, Xt = Xt)
summary(GAISLres)
plot(GAISLres, data = Xt)

# with suggestions
suggestions <- list(NULL, 250, c(250, 500), c(250, 625), c(250, 500, 750))
GAISLres <- cptgaisl(ObjFunc = arima_bic, N = Ts, suggestions = suggestions, XMat = XMatT, Xt = Xt)
summary(GAISLres)
plot(GAISLres, data = Xt)


## Multiple changepoint detection with model order selection

prange <- list(ar = c(0, 3), ma = c(0, 3))

# without suggestions
GAISLres <- cptgaisl(
  ObjFunc = arima_bic_order_pq, N = Ts, prange = prange,
  option = "both", XMat = XMatT, Xt = Xt
)
summary(GAISLres)
plot(GAISLres, data = Xt)

# with suggestions
suggestions <- list(NULL, 250, c(250, 500), c(250, 625), c(250, 500, 750))
GAISLres <- cptgaisl(
  ObjFunc = arima_bic_order_pq, N = Ts, prange = prange,
  suggestions = suggestions, option = "both", XMat = XMatT, Xt = Xt
)
summary(GAISLres)
plot(GAISLres, data = Xt)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/amoc_func.R
\name{amoc_mutation}
\alias{amoc_mutation}
\title{Jump mutation operator to produce offspring for AMOC problem}
\usage{
amoc_mutation(
  child,
  prange = NULL,
  minDist,
  pchangepoint = NULL,
  lmax = NULL,
  mmax = NULL,
  N = NULL
)
}
\arguments{
\item{child}{The child chromosome resulting from the \code{crossover} genetic
operator.}

\item{prange}{The default value is \code{NULL}. If there is no requirement
on model order selection, such an auxiliary argument is needed for \code{GA}
and \code{IslandGA} functions.}

\item{minDist}{The minimum length between two adjacent changepoints in
\code{\link{amoc_selection}} operator, which is also the jump magnitude in the
\code{amoc_mutation} operator.}

\item{pchangepoint}{An auxiliary argument is needed for \code{GA}
and \code{IslandGA} functions.}

\item{lmax}{An auxiliary argument is needed for \code{GA} and \code{IslandGA}
functions.}

\item{mmax}{An auxiliary argument is needed for \code{GA} and \code{IslandGA}
functions.}

\item{N}{An auxiliary argument is needed for \code{GA} and \code{IslandGA}
functions.}
}
\value{
The resulting child chromosome representation.
}
\description{
In a certain probability, the \code{mutation} genetic operator can be applied
to generate a new \code{child}. In this AMOC mutation operator, the new child
changepoint location can be down via a "jump" method. The child changepoint
location will jump \code{minDist} time units either to the left or right to
produce the changepoint location for the mutated child. The jump direction is
randomly decided with 0.5 probability.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cptga-class.R
\name{print.summary.cptga}
\alias{print.summary.cptga}
\title{Print Summary for a `cptga` Object}
\usage{
\method{print}{summary.cptga}(x, digits = getOption("digits"), max_display = 5, ...)
}
\arguments{
\item{x}{An object of class \code{cptga}, typically produced by a GA-based changepoint detection routine.}

\item{digits}{Number of digits to print for probabilities and fitness. Default taken from \code{getOption("digits")}.}

\item{max_display}{Maximum number of suggested solutions to display if suggestions are provided.}

\item{...}{Additional arguments (currently not used).}
}
\value{
Invisibly returns \code{NULL}. Called for its side effect of printing to the console.
}
\description{
Displays key information about the settings and results from a changepoint detection
procedure using the Genetic Algorithm (GA) stored in a `cptga` object. This includes
the algorithm configuration, population settings, optimization mode, and final
solution such as the number and location of changepoints and model parameters (if applicable).
}
\details{
When the GA is run in \code{option = "cp"} mode, only changepoint locations are shown.
If \code{option = "both"}, the output includes the selected model hyperparameters along
with changepoint locations.

The function uses plain text output to print a formatted summary to the console. If
\code{x@suggestions} is provided, only up to \code{max_display} suggestions will be shown.
}
\seealso{
\code{\link{cptga}}, \code{\link[=summary.cptga]{summary}}, \code{\link{plot.cptga}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ts_sim.R
\name{ts_sim}
\alias{ts_sim}
\title{Time series simulation with changepoint effects}
\usage{
ts_sim(
  Ts,
  beta,
  XMat,
  sigma,
  phi = NULL,
  theta = NULL,
  d = 0,
  Delta = NULL,
  CpLoc = NULL,
  seed = NULL
)
}
\arguments{
\item{Ts}{A integer indicating simulated time series length (sample size).}

\item{beta}{A parameter vector contains other mean function parameters without changepoint parameters.}

\item{XMat}{The covairates for time series mean function without changepoint indicators.}

\item{sigma}{The standard deviation for time series residuals \eqn{\epsilon_{t}}.}

\item{phi}{A vector for the autoregressive (AR) parameters for AR part.}

\item{theta}{A vector for the moving average (MA) parameters for MA part.}

\item{d}{A nonnegative integer for the order of differencing in the ARIMA model. Default is 0, which reduces to an ARMA model.}

\item{Delta}{The parameter vector contains the changepoint parameters for time series mean function.}

\item{CpLoc}{A vector contains the changepoint locations range from \eqn{1\leq\tau\leq T_{s}}.}

\item{seed}{The random seed for simulation reproducibility.}
}
\value{
The simulated time series with attributes:
\itemize{
  \item \code{DesignX}: The covariates including changepoint indicators.
  \item \code{mu}: The mean vector for the simulated time series.
  \item \code{theta}: The true parameter vector including changepoint effects.
  \item \code{CpEff}: The true changepoint magnitudes.
  \item \code{CpLoc}: The true changepoint locations.
  \item \code{arEff}: The AR coefficients.
  \item \code{maEff}: The MA coefficients.
  \item \code{diffEff}: The differencing order.
  \item \code{seed}: The random seed used for simulation.
}
}
\description{
This is a function to simulate time series with changepoint effects. See
details below.
}
\details{
The simulated time series \eqn{Z_t, t = 1, \ldots, T_s} is generated from
\deqn{Z_t = \mu_t + e_t.}

The error process \eqn{e_t} can be:
\itemize{
  \item IID Gaussian noise when \code{phi = NULL}, \code{theta = NULL},
    and \code{d = 0}.
  \item An ARMA(p,q) process when \code{d = 0} and at least one of
    \code{phi} or \code{theta} is specified.
  \item An ARIMA(p,d,q) process when \code{d > 0}.
}

The mean structure \eqn{\mu_t} is specified through \code{XMat} and
\code{beta}, and changepoint effects can be introduced as
\deqn{\mu_t = X_t^\top \beta + \Delta_1 I(t \ge \tau_1) + \cdots +
\Delta_m I(t \ge \tau_m),}
where \eqn{1 \le \tau_1 < \cdots < \tau_m \le T_s} are changepoint
locations and \eqn{\Delta_1, \ldots, \Delta_m} are changepoint magnitudes.
}
\examples{
##### M1: Time series observations are IID
Ts <- 1000
betaT <- c(0.5)
XMatT <- matrix(1, nrow = Ts, ncol = 1)
colnames(XMatT) <- "intercept"
sigmaT <- 1
DeltaT <- c(2, -2)
Cpprop <- c(1 / 4, 3 / 4)
CpLocT <- floor(Ts * Cpprop)

myts <- ts_sim(
  Ts = Ts, beta = betaT, XMat = XMatT, sigma = sigmaT,
  Delta = DeltaT, CpLoc = CpLocT, seed = 1234
)

##### M2: ARMA(2,1) model with constant mean
Ts <- 1000
betaT <- c(0.5)
XMatT <- matrix(1, nrow = Ts, ncol = 1)
colnames(XMatT) <- "intercept"
sigmaT <- 1
phiT <- c(0.5, -0.5)
thetaT <- c(0.8)
DeltaT <- c(2, -2)
CpLocT <- floor(Ts * c(1 / 4, 3 / 4))

myts <- ts_sim(
  Ts = Ts, beta = betaT, XMat = XMatT, sigma = sigmaT,
  phi = phiT, theta = thetaT, d = 0,
  Delta = DeltaT, CpLoc = CpLocT, seed = 1234
)

##### M3: ARIMA(1,1,1) model with changepoint effects
Ts <- 1000
betaT <- c(0.5)
XMatT <- matrix(1, nrow = Ts, ncol = 1)
colnames(XMatT) <- "intercept"
sigmaT <- 1
phiT <- c(0.5)
thetaT <- c(-0.3)
DeltaT <- c(2, -2)
CpLocT <- floor(Ts * c(1 / 4, 3 / 4))

myts <- ts_sim(
  Ts = Ts, beta = betaT, XMat = XMatT, sigma = sigmaT,
  phi = phiT, theta = thetaT, d = 1,
  Delta = DeltaT, CpLoc = CpLocT, seed = 1234
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/amoc_func.R
\name{amoc_crossover}
\alias{amoc_crossover}
\title{Average crossover operator to produce offspring for AMOC problem}
\usage{
amoc_crossover(mom, dad, prange = NULL, minDist, lmax, N)
}
\arguments{
\item{mom}{Among two selected individuals, \code{mom} represents the selected
chromosome representation with lower fitness function value.}

\item{dad}{Among two selected individuals, \code{dad} represents the selected
chromosome representation with larger fitness function value.}

\item{prange}{The default value is \code{NULL}. If there is no requirement
on model order selection, such an auxiliary argument is needed for \code{GA}
and \code{IslandGA} functions.}

\item{minDist}{The minimum length between two adjacent changepoints.}

\item{lmax}{The maximum possible length of the chromosome representation.}

\item{N}{The length of time series.}
}
\value{
The child chromosome that produced from \code{mom} and \code{dad} for
next generation.
}
\description{
In this crossover operator designed for AMOC problem, the new child is
produced by taking the average of the changepoint locations from \code{dad}
and \code{mom} and round to an integer. Note, every chromosome has at most
one candidate changepoint location.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\docType{package}
\name{changepointGA-package}
\alias{changepointGA}
\alias{changepointGA-package}
\title{changepointGA: Genetic Algorithms for Changepoint Models}
\description{
Tools for fitting changepoint models with genetic algorithms and Armadillo-backed linear algebra.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/mli171/changepointGA}
  \item Report bugs at \url{https://github.com/mli171/changepointGA/issues}
}

}
\author{
\strong{Maintainer}: Mo Li \email{mo.li@louisiana.edu}

Authors:
\itemize{
  \item Mo Li \email{mo.li@louisiana.edu}
  \item QiQi Lu \email{qlu2@vcu.edu}
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GAfunc.R
\name{mutation}
\alias{mutation}
\title{The default mutation operator in genetic algorithm}
\usage{
mutation(child, prange = NULL, minDist, pchangepoint, lmax, mmax, N)
}
\arguments{
\item{child}{The child chromosome resulting from the \code{crossover} genetic
operator.}

\item{prange}{Default is \code{NULL} for only changepoint detection. If
\code{prange} is specified as a list object, which contains the range of
each model order parameters for order selection (integers). The number of
order parameters must be equal to the length of \code{prange}.}

\item{minDist}{The required minimum distance between two adjacent changepoints.}

\item{pchangepoint}{The probability of changepoints for every time series.}

\item{lmax}{The user specified maximum number of changepoints, by default,
as \code{N/2 - 1}.}

\item{mmax}{The user specified maximum length of individual chromosome,
by default, as \code{2+N/2 + 1}.}

\item{N}{The sample size of the time series.}
}
\value{
The resulting child chromosome representation.
}
\description{
In a certain probability, the \code{mutation} genetic operator can be applied
to generate a new \code{child}. By default, the new child selection can be
down by the similar individual selection method in population initialization,
\code{\link{select_tau}}.
}
\details{
A function can apply mutation to the produced child with the specified
probability \code{pmutation} in \code{cptga} and
\code{cptgaisl}. If order selection is not requested
(\code{option = "cp"} in \code{cptga} and \code{cptgaisl}), the default
\code{\link{mutation}} operator function uses \code{select_tau} to select
a completely new individual with a new chromosome as the mutated child.
For details, see \code{\link{select_tau}}. If order selection is needed
(\code{option = "both"} in \code{cptga} and \code{cptgaisl}), we first
decide whether to keep the produced child's model order with a probability
of 0.5. If the child's model order is retained, the \code{select_tau}
function is used to select a completely new individual with a new chromosome
as the mutated child. If a new model order is selected from the candidate
model order set, there is a 0.5 probability to either select a completely new
individual with new changepoint locations or retain the original child's
changepoint locations for the mutated child. Note that the current model
orders in the child's chromosome are excluded from the set to avoid redundant
objective function evaluation. Finally, the function returns a vector
containing the modified chromosomes for mutated \code{child}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cptgaisl-class.R
\name{plot.cptgaisl}
\alias{plot.cptgaisl}
\title{Plot Time Series with Detected Changepoints from a `cptgaisl` Object}
\usage{
\method{plot}{cptgaisl}(
  x,
  data,
  show_segmean = TRUE,
  main = NULL,
  XTickLab = NULL,
  XTickPos = NULL,
  XAxisLab = "Time",
  YAxisLab = "Data",
  cex.lab = 1.3,
  cex.axis = 1.3,
  cex.main = 1.3,
  lwd = 2,
  ...
)
}
\arguments{
\item{x}{An object of class \code{cptgaisl}, typically returned by an island model genetic algorithm based
changepoint detection procedure.}

\item{data}{A numeric vector representing the observed univariate time series.}

\item{show_segmean}{Binary, whether to include the segments' means.}

\item{main}{Optional main title for the plot.}

\item{XTickLab}{Optional vector (e.g., numeric or date) for custom x-axis labels.
Must be the same length as \code{data}.}

\item{XTickPos}{Optional vector specifying which elements of \code{XTickLab} to show as ticks.}

\item{XAxisLab}{Optional label for the x-axis. Default is \code{"Time"}.}

\item{YAxisLab}{Optional label for the y-axis. Default is \code{"Data"}.}

\item{cex.lab}{Text size for axis labels and margin text. Default is \code{1.3}.}

\item{cex.axis}{Text size for axis tick labels. Default is \code{1.3}.}

\item{cex.main}{Text size for the main title. Default is \code{1.3}.}

\item{lwd}{Line width for vertical and horizontal dashed lines. Default is \code{2}.}

\item{...}{Additional graphical parameters passed to \code{plot()}.}
}
\value{
This function is called for its side effects and returns \code{NULL} invisibly.
}
\description{
This function visualizes a univariate time series along with the changepoints
identified by a island model genetic algorithm, as represented by a `cptgaisl` object.
Vertical dashed lines mark changepoint locations, and segment means are shown as horizontal
dashed lines. The optimal fitness value and changepoint locations are
displayed as margin text.
}
\details{
If \code{XTickLab} is supplied and matches the length of \code{data}, it is used for
the x-axis; otherwise, the default sequence \code{1:length(data)} is used.

If the genetic algorithm was run with \code{option = "both"}, the function skips hyperparameters
in the chromosome when extracting changepoint positions.

The plot displays vertical dashed lines at changepoint locations and horizontal dashed
lines for the mean of each segment. Fitness and changepoint summaries are shown above the plot.
}
\seealso{
\code{\link[=summary,cptgaisl-method]{summary}}, \code{\link{print.summary.cptgaisl}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cpt_dist.R
\name{cpt_dist}
\alias{cpt_dist}
\title{Comparing multiple changepoint configurations by pairwise distance}
\usage{
cpt_dist(tau1, tau2, N)
}
\arguments{
\item{tau1}{A vector contains the changepoint locations for \eqn{C_{1}}
for comparison. A value \code{NULL} is required if there is no changepoint detected.}

\item{tau2}{A vector contains the changepoint locations for  \eqn{C_{2}}
for comparison. A value \code{NULL} is required if there is no changepoint detected.}

\item{N}{The simulated time series sample size. Two changepoint configurations
should have the same \eqn{n} values.}
}
\value{
\item{dist}{The calculated distance.}
}
\description{
This function is to calculate the distance metric (Shi et al., 2022) between
two changepoint configurations, \eqn{C_{1}=\{\tau_{1}, \ldots, \tau_{m}\}}
and \eqn{C_{2}=\{\eta_{1}, \ldots, \eta_{k}\}}, where \eqn{\tau}'s are the
changepoint locations.
}
\details{
The pairwise distance was proposed by Shi et al. (2022),
\deqn{d(C_{1}, C_{2})=|m-k|+ \min(A(C_{1}, C_{2})),}
where \eqn{m} is the number of changepoints in configuration \eqn{C_{1}} and
\eqn{k} is the number of changepoints in configuration \eqn{C_{2}}.
The term \eqn{\min(A(C_{1}, C_{2}))} reflects the cost of matching
changepoint locations between \eqn{C_{1}} and \eqn{C_{2}} and can be calculated
using the linear assignment method. Details can be found in Shi et al. (2022).
Note: if one configuration doesn't contain any changepoints (valued \code{NULL}),
the distance is defined as \eqn{|m-k|}. The function can also be used
to examine changepoint detection performance in simulation studies. Given the
true changepoint configuration, \eqn{C_{true}}, used in generating the time
series, the calculated distance between the estimated multiple changepoint
configuration, \eqn{C_{est}=\{\eta_{1}, \ldots, \eta_{k}\}}, and \eqn{C_{true}}
can be used to evaluate the performance of the detection algorithm.
}
\examples{
N <- 100

# both tau1 and tau2 has detected changepoints
tau2 <- c(25, 50, 75)
tau1 <- c(20, 35, 70, 80, 90)
cpt_dist(tau1 = tau1, tau2 = tau2, N = N)

# either tau1 or tau2 has zero detected changepoints
cpt_dist(tau1 = tau1, tau2 = NULL, N = N)
cpt_dist(tau1 = NULL, tau2 = tau2, N = N)

# both tau1 and tau2 has zero detected changepoints
cpt_dist(tau1 = NULL, tau2 = NULL, N = N)
}
\references{
{
  Shi, X., Gallagher, C., Lund, R., & Killick, R. (2022).
  A comparison of single and multiple changepoint techniques for time series data.
  \emph{Computational Statistics & Data Analysis}, 170, 107433.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/amoc_func.R
\name{amoc_selection}
\alias{amoc_selection}
\title{The parents selection genetic algorithm operator for AMOC problem}
\usage{
amoc_selection(pop, popFit)
}
\arguments{
\item{pop}{A matrix contains the chromosomes for all individuals. The number of
rows is equal to \code{lmax} and the number of columns is equal to the
\code{popSize}.}

\item{popFit}{A vector contains the objective function value (population fit)
being associated to each individual chromosome from above.}
}
\value{
A list contains the chromosomes for \code{dad} and \code{mom}.
}
\description{
The genetic algorithm require to select a pair of chromosomes, representing
\code{dad} and \code{mom}, for the \code{crossover} operator to
produce offspring (individual for next generation). Here, the same linear
ranking method in \code{\link{selection_linear_rank}} is used to select a pair
of chromosomes for \code{dad} and \code{mom} in the at most one changepoint
(AMOC) problem. By default, the dad has better fit/smaller fitness function
value/larger rank than \code{mom}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{uniform_crossover}
\alias{uniform_crossover}
\title{Uniform crossover to produce offsprings}
\usage{
uniform_crossover(mom, dad, prange, minDist, lmax, N)
}
\arguments{
\item{mom}{Among two selected individuals, \code{mom} represents the selected
chromosome representation with lower fitness function value.}

\item{dad}{Among two selected individuals, \code{dad} represents the selected
chromosome representation with larger fitness function value.}

\item{prange}{Default value is \code{NULL} for only changepoint detection. If
\code{prange} is specified as a list object, which contains the range of
each model order parameters for order selection (integers). The number of
order parameters must be equal to the length of \code{prange}.}

\item{minDist}{The required minimum distance between two adjacent changepoints.}

\item{lmax}{The user specified maximum number of changepoints, by default,
as \code{N/2 - 1}.}

\item{N}{The length of time series.}
}
\value{
The child chromosome that produced from \code{mom} and \code{dad} for
next generation.
}
\description{
In uniform crossover, typically, each bit is chosen from either parent with
equal probability. Other mixing ratios are sometimes used, resulting in
offspring which inherit more genetic information from one parent than the
other. In a uniform crossover, we don’t divide the chromosome into segments,
rather we treat each gene separately. In this, we essentially flip a coin
for each chromosome to decide whether or not it will be included in the
off-spring. If model order selection is requested, each child's model order
has the equal probability (0.5) from \code{dad} and \code{mom}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cptgaisl-class.R
\docType{class}
\name{cptgaisl-class}
\alias{cptgaisl-class}
\alias{summary,cptgaisl-method}
\alias{summary.cptgaisl}
\title{S4 Class Definition for `cptgaisl`}
\usage{
\S4method{summary}{cptgaisl}(object, ...)
}
\arguments{
\item{object}{An object of class \code{cptgaisl}.}

\item{...}{Additional arguments (ignored).}
}
\value{
An object of class \code{cptgaisl}
}
\description{
S4 Class for Island Model Genetic Algorithm-Based Changepoint Detection
}
\details{
This class stores results and settings for the island model genetic algorithm (`cptgaisl`)
used in multiple changepoint detection and optional model order selection.
}
\section{Slots}{

\describe{
\item{\code{call}}{The matched call that created the object.}

\item{\code{N}}{The sample size of the time series.}

\item{\code{prange}}{A list or NULL. Ranges for each model order parameter when \code{option = "both"}.}

\item{\code{popSize}}{Integer. It represents the total number of individuals in each generation, which equal to the number of islands multiplied by the size of each island (i.e., \code{popSize = numIslands × Islandsize}).}

\item{\code{numIslands}}{Integer. The number of islands (sub-populations).}

\item{\code{Islandsize}}{Numerical value. The number of individuals in each island.}

\item{\code{pcrossover}}{Probability of crossover operation.}

\item{\code{pmutation}}{Probability of mutation operation.}

\item{\code{pchangepoint}}{Prior probability of changepoint occurrence.}

\item{\code{minDist}}{Minimum distance between adjacent changepoints.}

\item{\code{mmax}}{Maximum number of changepoints allowed.}

\item{\code{lmax}}{Maximum length of chromosome.}

\item{\code{maxMig}}{Maximum number of migrations allowed.}

\item{\code{maxgen}}{Maximum number of generations per island before migration.}

\item{\code{maxconv}}{Number of migrations with no improvement before stopping.}

\item{\code{option}}{Either "cp" or "both".}

\item{\code{monitoring}}{Logical. If TRUE, prints intermediate output.}

\item{\code{parallel}}{Logical. Whether parallel computation is used.}

\item{\code{nCore}}{Integer or NULL. Number of cores for parallelization.}

\item{\code{tol}}{Tolerance threshold for fitness improvement.}

\item{\code{seed}}{Integer or NULL. Seed for reproducibility.}

\item{\code{suggestions}}{A list or NULL. Suggested changepoint configurations.}

\item{\code{Island}}{A 3D array storing all individual chromosomes in current generation across islands. Dimensions are \code{lmax} × \code{Islandsize} × \code{numIslands}, representing chromosome length, individuals per island, and number of islands, respectively.}

\item{\code{IslandFit}}{A matrix of fitness values in current generation with dimensions \code{Islandsize} × \code{numIslands}, where each column corresponds to one island's population.}

\item{\code{overbestchrom}}{A vector. The best chromosome ever found.}

\item{\code{overbestfit}}{Numeric. The best fitness score obtained.}

\item{\code{bestfit}}{Numeric vector recording best fitness per migration.}

\item{\code{countMig}}{Integer vector tracking number of migrations.}

\item{\code{count}}{Integer vector tracking total generations.}

\item{\code{convg}}{Integer vector for convergence diagnostics.}
}}

\seealso{
\code{\link{cptgaisl}}, \code{\link{cptgaisl-class}}, \code{\link{random_population}}, \code{\link{selection_linear_rank}}, \code{\link{uniform_crossover}}, \code{\link{mutation}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{random_population}
\alias{random_population}
\title{Random population initialization}
\usage{
random_population(popSize, prange, N, minDist, pchangepoint, mmax, lmax)
}
\arguments{
\item{popSize}{An integer represents the number of individual in each
population for GA (or subpopulation for IslandGA).}

\item{prange}{Default is \code{NULL} for only changepoint detection. If
\code{prange} is specified as a list object, which contains the range of
each model order parameters for order selection (integers). The number of
order parameters must be equal to the length of \code{prange}.}

\item{N}{The length of time series.}

\item{minDist}{The minimum length between two adjacent changepoints.}

\item{pchangepoint}{Same as \code{pchangepoint} from \code{\link{cptga}} or 
\code{\link{cptgaisl}}, the probability that a changepoint has occurred.}

\item{mmax}{The maximum possible number of changepoints in the data set.}

\item{lmax}{The maximum possible length of the chromosome representation.}
}
\value{
A matrix that contains each individual's chromosome.
}
\description{
Randomly generate the individuals' chromosomes (changepoint confirgurations)
to construct the first generation population.
}
\details{
Each population can be stored in a matrix with \code{lmax} rows and 
\code{popSize} columns, where each column represents an individual chromosome
in the format of
\deqn{C = (m, \boldsymbol{s}, \boldsymbol{\tau}, N+1)',} 
in \code{\link{cptga}} or \code{\link{cptgaisl}}. This function can randomly 
initialize the population matrix with some imposed constraints, such as 
ensuring the number of time points between two adjacent changepoints is 
greater than or equal to \code{minDist}. This prevents unrealistic scenario 
of two changepoints being too close to each other and helps reduce the total 
number of admissible solutions in the search space. Users can adjust the 
level of searching space reduction by setting an appropriate value for 
\code{minDist}. During population initialization, each changepoint location 
in \eqn{\boldsymbol{\tau}} can be selected sequentially. With a specified 
probability \code{pchangepoint} denoting the probability of a time point 
being selected as a changepoint, the first changepoint \eqn{\tau_{1}} is 
randomly picked between \eqn{t=1+minDist} and \eqn{t=N}. Then \eqn{\tau_{2}} 
is randomly selected from a smaller range between \eqn{t=\tau_{1}+minDist} 
and \eqn{t=N}. The process is continued until the last admissible changepoint 
\eqn{t=N-minDist} is exceeded, and the number of changepoints \eqn{m} is 
obtained automatically. For added flexibility, users can specify their own 
population initialization function. The default population initialization uses 
\code{\link{select_tau}} to select the chromosome for the first generation 
population.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{select_tau}
\alias{select_tau}
\title{Randomly select the chromosome}
\usage{
select_tau(N, prange, minDist, pchangepoint, mmax, lmax)
}
\arguments{
\item{N}{The length of time series.}

\item{prange}{A list object containing the possible range for other
pre-defined model parameters, i.e. AR/MA order of ARMA models.}

\item{minDist}{The minimum length between two adjacent changepoints.}

\item{pchangepoint}{Same as \code{Pchangepoint}, the probability that a changepoint has occurred.}

\item{mmax}{The maximum possible number of changepoints in the data set.}

\item{lmax}{The maximum possible length of the chromosome representation.}
}
\value{
A single changepoint configuration format as above.
}
\description{
Randomly select the changepoint configuration for population initialization.
The selected changepoint configuration represents a changepoint chromosome.
The first element of the chromosome represent the number of changepoints
and the last non-zero element always equal to the length of time series + 1.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/arima_bic_order_pq.R
\name{arima_bic_order_pq}
\alias{arima_bic_order_pq}
\title{Calculating BIC for Multiple changepoint detection with AR and MA order selection}
\usage{
arima_bic_order_pq(chromosome, plen = 2, XMat, Xt)
}
\arguments{
\item{chromosome}{A vector consists of the number of changepoints, the order
of AR component (refers to the number of lagged terms used to model the
current value of a time series), the order of MA component (refers to the
number of lagged error terms used to model the current value of a time
series), the changepoint locations, and a value of time series sample size
plus 1 ($N+1$) indicating the end of the chromosome.}

\item{plen}{The number of model order parameters that need to be selected.
If model order selection needs to be performed simultaneously with the
changepoint detection task, \code{plen} should be nonzero.}

\item{XMat}{A matrix contains the covariates, but not includes changepoint
effects, for time series regression.}

\item{Xt}{The simulated ARMA time series from \code{ts.sim} function.}
}
\value{
The BIC value of the objective function.
}
\description{
The objective function for changepoint search in Autoregressive
moving average with model order selection via Bayesian Information Criterion
(BIC).
}
\examples{
Ts <- 1000
XMatT <- matrix(1, nrow = Ts, ncol = 1)
Xt <- ts_sim(
  Ts = Ts, beta = 0.5, XMat = XMatT, sigma = 1, phi = 0.5, theta = 0.8,
  Delta = c(2, -2), CpLoc = c(250, 750), seed = 1234
)

# one chromosome representation
chromosome <- c(2, 1, 1, 250, 750, 1001)
arima_bic_order_pq(chromosome, plen = 2, XMat = XMatT, Xt = Xt)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cptga-class.R
\docType{class}
\name{cptga-class}
\alias{cptga-class}
\alias{summary,cptga-method}
\alias{summary.cptga}
\title{S4 Class Definition for `cptga`}
\usage{
\S4method{summary}{cptga}(object, ...)
}
\arguments{
\item{object}{An object of class \code{cptga}.}

\item{...}{Additional arguments (ignored).}
}
\value{
An object of class \code{cptga}.
}
\description{
S4 Class for Genetic Algorithm-Based Changepoint Detection
}
\details{
An object of class \code{cptga} stores results and configuration settings
for changepoint detection using a Genetic Algorithm (GA), optionally with
simultaneous model order selection. This class records GA control parameters,
intermediate population structures, and the optimal solution found.
}
\section{Slots}{

\describe{
\item{\code{call}}{The matched call that created the object.}

\item{\code{N}}{The sample size of the time series.}

\item{\code{prange}}{A list object. Default is \code{NULL}. If specified, it contains the
ranges for each model order parameter (integers). Required when \code{option = "both"}
is used for joint changepoint and model selection.}

\item{\code{popSize}}{An integer representing the number of individuals in each GA population.}

\item{\code{pcrossover}}{The probability that the crossover operator is applied to two chromosomes.}

\item{\code{pmutation}}{The probability that the mutation operator is applied to a chromosome.}

\item{\code{pchangepoint}}{The prior probability that a changepoint has occurred at each location.}

\item{\code{minDist}}{The minimum allowed distance between two adjacent changepoints.}

\item{\code{mmax}}{The maximum possible number of changepoints. Typically set based on time series length and \code{option}.}

\item{\code{lmax}}{The maximum length of the chromosome. Typically set based on time series length and \code{option}.}

\item{\code{maxgen}}{The maximum number of generations the GA is allowed to run.}

\item{\code{maxconv}}{If the optimal fitness value does not improve over this many generations, GA stops.}

\item{\code{option}}{A character string: either \code{"cp"} for changepoint detection only, or \code{"both"} for changepoint detection and model order selection.}

\item{\code{monitoring}}{Logical. If \code{TRUE}, prints intermediate GA progress.}

\item{\code{parallel}}{Logical. If \code{TRUE}, enables parallel computation for fitness evaluation.}

\item{\code{nCore}}{Integer or \code{NULL}. Number of cores used for parallel computation when \code{parallel = TRUE}.}

\item{\code{tol}}{Numeric. Tolerance for determining GA convergence. Default is \code{1e-5}.}

\item{\code{seed}}{An integer or \code{NULL}. Random seed for reproducibility.}

\item{\code{suggestions}}{A list or \code{NULL}. Each element provides suggested changepoint locations
to guide initial population design and potentially accelerate convergence.}

\item{\code{population}}{A matrix where each row represents an individual chromosome in the current population.}

\item{\code{fitness}}{A numeric vector containing the fitness values of individuals in the current generation.}

\item{\code{overbestchrom}}{A vector representing the best chromosome found over all generations.}

\item{\code{overbestfit}}{A numeric scalar. The best (smallest) fitness value achieved.}

\item{\code{bestfit}}{A numeric vector recording the best fitness value in each generation.}

\item{\code{count}}{A numeric value indicating the number of generations the GA actually ran.}

\item{\code{convg}}{A numeric vector representing convergence information. A value of \code{0} indicates the algorithm successful completion. A value of \code{1} indicates the the total number of generations exceeds the pre-specified \code{maxgen} limit.}
}}

\seealso{
\code{\link{cptga}}, \code{\link{cptga-class}}, \code{\link{random_population}}, \code{\link{selection_linear_rank}}, \code{\link{uniform_crossover}}, \code{\link{mutation}}.
}
